# DF2DBPlus in this code the dense dataframe is converting databases into different transactional, temporal, utility types.
#
# **Importing this algorithm into a python program**
# --------------------------------------------------------
#
#             from PAMI.extras.DF2DB import DF2DBPlus as dfdbp
#
#             obj = dfdbp.DF2DBPlus(idf, ">=", 16)
#
#             obj.getTransactional("outputFileName") # To create a transactional database
#
#             obj.getTDB("outputFileName")   # To create a temporal database
#
#             obj.getUDB("outputFileName")    # To create a utility database
#




__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from PAMI.extras.DF2DB.DenseFormatDFPlus import *
from PAMI.extras.DF2DB.SparseFormatDFPlus import *
import sys


class DF2DBPlus:
    """
    :Description:  This class create database from DataFrame. Threshold values and conditions are defined to each item by thresholdConditonDF.

    :Attributes:
        :param inputDF: DataFrame :
             It is sparse or dense DataFrame
        :param thresholdConditionDF: pandas.DataFrame :
            It is DataFrame to contain threshold values and condition each item
        :param condition: str :
             It is condition of all item
        :param DFtype: str :
             It is DataFrame type. It should be sparse or dense. Default DF is sparse.



    **Importing this algorithm into a python program**
    --------------------------------------------------------
    .. code-block:: python

             from PAMI.extras.DF2DB import DF2DBPlus as dfdbp

             obj = dfdbp.DF2DBPlus(idf, ">=", 16)

             obj.getTransactional("outputFileName") # To create a transactional database

             obj.getTDB("outputFileName")   # To create a temporal database

             obj.getUDB("outputFileName")    # To craete a utility database
    """

    def __init__(self, inputDF, thresholdConditionDF, DFtype='sparse') -> None:
        self.inputDF = inputDF
        self.thresholdConditionDF = thresholdConditionDF
        self.DFtype = DFtype.lower()
        if DFtype == 'sparse':
            self.DF2DB = SparseFormatDFPlus(self.inputDF, self.thresholdConditionDF)
        elif DFtype == 'dense':
            self.DF2DB = DenseFormatDFPlus(self.inputDF, self.thresholdConditionDF)
        else:
            raise Exception('DF type should be sparse or dense')

    def getTransactional(self, outputFile) -> str:
        """
        create transactional database and return outputFileName
        :param outputFile: file name or path to store database
        :type outputFile: str
        :return: outputFile name
        :rtype: str
        """
        self.DF2DB.createTransactional(outputFile)
        return self.DF2DB.getFileName()

    def getTDB(self, outputFile) -> str:
        """
        create temporal database and return outputFile name
        :param outputFile: file name or path to store database
        :type outputFile: str
        :return: outputFile name
        :rtype: str
        """
        self.DF2DB.createTemporal(outputFile)
        return self.DF2DB.getFileName()

    def getUDB(self, outputFile) -> str:
        """
        create utility database and return outputFile name
        :param outputFile:  file name or path to store database
        :type outputFile: str
        :return: outputFile name
        :rtype: str
        """
        self.DF2DB.createUtility(outputFile)
        return self.DF2DB.getFileName()
if __name__ == '__main__':
    obj = DF2DBPlus(sys.argv[1], sys.argv[2], sys.argv[3])
    obj.getTransactional(sys.argv[4])


# DF2Tex is used to convert the given dataframe into LatexFile.
#
# **Importing this algorithm into a python program**
#
#             from PAMI.extras.graph import DF2Tex
#
#             obj = DF2Tex()
#
#             obj.generateLatexCode(result, "minSup", "runtime", "algorithmColumn")
#
#             obj.printLatex()
#
#             obj.save("outputFile.tex")
#

__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import pandas as pd

class DF2Tex:
    """
    **About this algorithm**

    :**Description**:  generateLatexFileFromDataFrame is used to convert the given DataFrame into a LaTeX file.

    :**Reference**:

    **Attributes:** - **latexCode** (*str*) -- Stores the generated LaTeX code.

    **Methods:**    - **generateLatexCode(result: pd.DataFrame, xColumn, yColumn, algorithmColumn=None)** -- *Generates LaTeX code based on DataFrame columns.*

                    - **print_latex()** -- *Prints the LaTeX code.*

                    - **save(outputFileName: str)** -- *Saves the LaTeX code to a .tex file.*

    **Importing this algorithm into a Python program**

    .. code-block:: python

            from PAMI.extras.graph import DF2Tex

            result = Dataframe

            obj = DF2Tex()

            obj.generateLatexCode(result, "minSup", "runtime", "algorithmColumn")

            obj.printLatex()

            obj.save("outputFile.tex")
    """

    def __init__(self):
        self.latexCode = ""

    def generateLatexCode(self, result: pd.DataFrame, xColumn, yColumn, algorithmColumn=None) -> None:
        """
        Generate LaTeX code from a given DataFrame.
        """
        # Check if DataFrame is empty
        if result.empty:
            raise ValueError("The input DataFrame is empty. Please provide a DataFrame with data.")

        # Check if required columns exist
        requiredColumns = [xColumn, yColumn]
        if algorithmColumn:
            requiredColumns.append(algorithmColumn)
        for col in requiredColumns:
            if col not in result.columns:
                raise ValueError(f"Column '{col}' does not exist in the DataFrame.")

        titles = [xColumn, yColumn]
        legendary = pd.unique(result[algorithmColumn] if algorithmColumn else result.iloc[:, 0].values.ravel())
        xaxisValues = result[xColumn].values
        yaxisValues = result[yColumn].values
        algo = result[algorithmColumn].values if algorithmColumn else result.iloc[:, 0].values
        xLabel = titles[0]
        colors = ['red', 'blue', 'green', 'black', 'yellow']
        latexCode = ""
        latexCode += "\\begin{axis}[\n\txlabel={\\Huge{" + xLabel + "}},"
        latexCode += "\n\tylabel={\\Huge{" + titles[1] + "}},"
        latexCode += "\n\txmin=" + str(min(xaxisValues)) + ", xmax=" + str(max(xaxisValues)) + ",]\n"

        for num, legend in enumerate(legendary):
            color = colors[num % len(colors)]  # Ensure color index is within range
            latexCode += "\\addplot+  [" + color + "]\n\tcoordinates {\n"
            for x, y, a in zip(xaxisValues, yaxisValues, algo):
                if legend == a:
                    latexCode += f"({x},{y})\n"
            latexCode += "\t};   \\addlegendentry{" + str(legend) + "}\n"
        latexCode += "\\end{axis}"
        self.latexCode = latexCode

    def printLatex(self):
        """
        Print the generated LaTeX code.
        """
        print(self.latexCode)

    def save(self, outputFileName):
        """
        Save the generated LaTeX code to a file.

        :param outputFileName: The name of the output .tex file.
        :raises ValueError: If LaTeX code has not been generated.

        """
        if not self.latexCode:
            raise ValueError("LaTeX code is empty. Please generate it before saving.")
        with open(outputFileName, "w") as latexwriter:
            latexwriter.write(self.latexCode)
        print(f"LaTeX file saved as {outputFileName}")

# Example usage
# result = pd.DataFrame()
# obj = DF2Tex()
# generateLatexCode function as four parameters dataFrame, xColumn-name, yColumn-name,
# algorithmColumn-name is optional
# obj.generateLatexCode(result, "minSup", "runtime", "algorithmColumn")
# printLatexCode function prints the output of latex file
# obj.printLatex()
# save function gives the outputFile
# obj.save("outputFile.tex")



# This code implements the CFPGrowthPlus algorithm for mining frequent patterns with multiple minimum support thresholds from a transactional dataset.
#
# **Importing this algorithm into a python program**
#
#             from PAMI.multipleMinimumSupportBasedFrequentPattern.basic import CFPGrowthPlus as alg
#
#             iFile = "sample.txt"
#
#             MIS = "MIS.txt"
#
#             obj = alg.CFPGrowthPlus(iFile, MIS, sep)
#
#             obj.startMine()
#
#             frequentPatterns = obj.getPatterns()
#
#             print("Total number of Frequent Patterns:", len(frequentPatterns))
#
#             obj.savePatterns(oFile)
#
#             Df = obj.getPatternInDataFrame()
#
#             memUSS = obj.getMemoryUSS()
#
#             print("Total Memory in USS:", memUSS)
#
#             memRSS = obj.getMemoryRSS()
#
#             print("Total Memory in RSS", memRSS)
#
#             run = obj.getRuntime()
#
#             print("Total ExecutionTime in seconds:", run)





__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from PAMI.multipleMinimumSupportBasedFrequentPattern.basic import abstract as _fp
from deprecated import deprecated

_fp._sys.setrecursionlimit(20000)
# MIS = {}

class _Node:
    """
    A class used to represent the node of frequentPatternTree

    :Attributes:

        itemId: int
            storing item of a node
        counter: int
            To maintain the support of node
        parent: node
            To maintain the parent of node
        children: list
            To maintain the children of node

    :Methods:

        addChild(node)
            Updates the nodes children list and parent for the given node

    """

    def __init__(self, item, children):
        self.itemId = item
        self.counter = 1
        self.parent = None
        self.children = children

    def addChild(self, node):
        """
        Retrieving the child from the tree

        :param node: Children node.
        :type node: Node
        :return: Updates the children nodes and parent nodes
        """
        self.children[node.itemId] = node
        node.parent = self


class _Tree:
    """
    A class used to represent the frequentPatternGrowth tree structure

    :Attributes:

        root : Node
            The first node of the tree set to Null.
        summaries : dictionary
            Stores the nodes itemId which shares same itemId
        info : dictionary
            frequency of items in the transactions

    :Methods:

        addTransaction(transaction, freq)
            adding items of  transactions into the tree as nodes and freq is the count of nodes
        getFinalConditionalPatterns(node)
            getting the conditional patterns from fp-tree for a node
        getConditionalPatterns(patterns, frequencies)
            sort the patterns by removing the items with lower minSup
        generatePatterns(prefix)
            generating the patterns from fp-tree
    """

    def __init__(self):
        self.root = _Node(None, {})
        self.summaries = {}
        self.info = {}

    def addTransaction(self, transaction, count):
        """
        adding transaction into tree

        :param transaction: it represents the one transactions in database
        :type transaction: list
        :param count: frequency of item
        :type count: int
        """

        # This method takes transaction as input and returns the tree
        currentNode = self.root
        for i in range(len(transaction)):
            if transaction[i] not in currentNode.children:
                newNode = _Node(transaction[i], {})
                newNode.freq = count
                currentNode.addChild(newNode)
                if transaction[i] in self.summaries:
                    self.summaries[transaction[i]].append(newNode)
                else:
                    self.summaries[transaction[i]] = [newNode]
                currentNode = newNode
            else:
                currentNode = currentNode.children[transaction[i]]
                currentNode.freq += count

    def getFinalConditionalPatterns(self, alpha, support):
        """
        generates the conditional patterns for a node

        :param alpha: node to generate conditional patterns
        :return: returns conditional patterns, frequency of each item in conditional patterns

        """
        finalPatterns = []
        finalFreq = []
        for i in self.summaries[alpha]:
            set1 = i.freq
            set2 = []
            while i.parent.itemId is not None:
                set2.append(i.parent.itemId)
                i = i.parent
            if len(set2) > 0:
                set2.reverse()
                finalPatterns.append(set2)
                finalFreq.append(set1)
        finalPatterns, finalFreq, info = self.getConditionalTransactions(finalPatterns, finalFreq, support)
        return finalPatterns, finalFreq, info

    @staticmethod
    def getConditionalTransactions(ConditionalPatterns, conditionalFreq, support):
        """
        To calculate the frequency of items in conditional patterns and sorting the patterns

        :param ConditionalPatterns: paths of a node
        :param conditionalFreq: frequency of each item in the path
        :return: conditional patterns and frequency of each item in transactions
        """
        #global _minSup
        pat = []
        freq = []
        data1 = {}
        for i in range(len(ConditionalPatterns)):
            for j in ConditionalPatterns[i]:
                if j in data1:
                    data1[j] += conditionalFreq[i]
                else:
                    data1[j] = conditionalFreq[i]
        up_dict = {k: v for k, v in data1.items() if v >= support}
        #up_dict = data1.copy()
        count = 0
        for p in ConditionalPatterns:
            p1 = [v for v in p if v in up_dict]
            trans = sorted(p1, key=lambda x: (up_dict.get(x), -x), reverse=True)
            if len(trans) > 0:
                pat.append(trans)
                freq.append(conditionalFreq[count])
            count += 1
        return pat, freq, up_dict

    def generatePatterns(self, prefix):
        """
        To generate the frequent patterns

        :param prefix: an empty list
        :return: Frequent patterns that are extracted from fp-tree

        """
        global minMIS, MIS
        for i in sorted(self.summaries, key=lambda x: (self.info.get(x), -x)):
            pattern = prefix[:]
            pattern.append(i)
            sup = []
            for j in pattern:
                sup.append(MIS[j])
            #if self.info[i] >= minMIS:
            if self.info[i] >= min(sup):
              yield pattern, self.info[i]
            #patterns, freq, info = self.getFinalConditionalPatterns(i, min(sup))
            patterns, freq, info = self.getFinalConditionalPatterns(i, minMIS)
            conditionalTree = _Tree()
            conditionalTree.info = info.copy()
            for pat in range(len(patterns)):
                conditionalTree.addTransaction(patterns[pat], freq[pat])
            if len(patterns) > 0:
                for q in conditionalTree.generatePatterns(pattern):
                    yield q

minMIS = 0
MIS = {}

class CFPGrowthPlus(_fp._frequentPatterns):
    """
    **About this algorithm**

    :**Description**: This code implements the CFPGrowthPlus algorithm for mining frequent patterns with multiple minimum support thresholds from a transactional dataset.

    :**Reference**:   R. Uday Kiran P. Krishna Reddy Novel techniques to reduce search space in multiple minimum supports-based frequent
                  pattern mining algorithms. 11-20 2011 EDBT https://doi.org/10.1145/1951365.1951370

    :**Parameters**:    - **iFile** (*str*) -- Name of the Input file to mine complete set of Uncertain Multiple Minimum Support Based Frequent patterns.
                        - **oFile** (*str*) -- Name of the output file to store complete set of Uncertain Minimum Support Based Frequent patterns.
                        - **MIS** (*str*) -- Name of the MIS file to mine complete set of Uncertain Multiple Minimum Support Based Frequent patterns.
                        - **sep** (*str*) -- This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.

     :**Attributes**:   - **startTime** (*float*) -- *To record the start time of the mining process.*
                        - **endTime** (*float*) -- *To record the completion time of the mining process.*
                        - **finalPatterns** (*dict*) -- *Storing the complete set of patterns in a dictionary variable.*
                        - **memoryUSS** (*float*) -- *To store the total amount of USS memory consumed by the program.*
                        - **memoryRSS** (*float*) -- *To store the total amount of RSS memory consumed by the program.*
                        - **Database** (*list*) -- *To store the transactions of a database in list.*
                        - **mapSupport** (*Dictionary*) -- *To maintain the information of item and their frequency.*
                        - **tree** (*class*) --  *it represents the Tree class.*

    :**Methods**:       - **startMine()** -- *Mining process will start from here.*
                        - **getPatterns()** -- *Complete set of patterns will be retrieved with this function.*
                        - **savePatterns(oFile)** -- *Complete set of frequent patterns will be loaded in to a output file.*
                        - **getPatternsAsDataFrame()** -- *Complete set of frequent patterns will be loaded in to a dataframe.*
                        - **getMemoryUSS()** -- *Total amount of USS memory consumed by the mining process will be retrieved from this function.*
                        - **getMemoryRSS()** -- *Total amount of RSS memory consumed by the mining process will be retrieved from this function.*
                        - **getRuntime()** -- *Total amount of runtime taken by the mining process will be retrieved from this function.*
                        - **creatingItemSets()** -- *Scans the dataset or dataframes and stores in list format.*
                        - **frequentOneItem()** -- *Extracts the one-frequent patterns from transactions.*

    **Execution methods**

    **Terminal command**

     .. code-block:: console

          Format:

          (.venv) $ python3 CFPGrowthPlus.py <inputFile> <outputFile> <MISFile>

          Example Usage:

          (.venv) $ python3 CFPGrowthPlus.py sampleDB.txt patterns.txt MISFile.txt

            .. note:: minSup  will be considered in support count or frequency


    **Calling from a python program**

    .. code-block:: python

            from PAMI.multipleMinimumSupportBasedFrequentPattern.basic import CFPGrowthPlus as alg

            iFile = "sample.txt"

            MIS = "MIS.txt"

            obj = alg.CFPGrowthPlus(iFile, MIS, sep)

            obj.startMine()

            frequentPatterns = obj.getPatterns()

            print("Total number of Frequent Patterns:", len(frequentPatterns))

            obj.savePatterns(oFile)

            Df = obj.getPatternInDataFrame()

            memUSS = obj.getMemoryUSS()

            print("Total Memory in USS:", memUSS)

            memRSS = obj.getMemoryRSS()

            print("Total Memory in RSS", memRSS)

            run = obj.getRuntime()

            print("Total ExecutionTime in seconds:", run)

    **Credits**

        The complete program was written by P.Likhitha  under the supervision of Professor Rage Uday Kiran.\n

    """

    __startTime = float()
    __endTime = float()
    _MIS = str
    __finalPatterns = {}
    _iFile = " "
    _oFile = " "
    _sep = " "
    __memoryUSS = float()
    __memoryRSS = float()
    __Database = []
    __mapSupport = {}
    __lno = 0
    __tree = _Tree()
    __rank = {}
    __rankDup = {}

    def __init__(self, iFile, MIS, sep='\t'):
        super().__init__(iFile, MIS, sep)

    def __creatingItemSets(self):
        """
        Storing the complete transactions of the database/input file in a database variable

        """
        self.__Database = []
        if isinstance(self._iFile, _fp._pd.DataFrame):
            if self._iFile.empty:
                print("its empty..")
            i = self._iFile.columns.values.tolist()
            if 'Transactions' in i:
                self.__Database = self._iFile['Transactions'].tolist()

            # print(self.Database)
        if isinstance(self._iFile, str):
            if _fp._validators.url(self._iFile):
                data = _fp._urlopen(self._iFile)
                for line in data:
                    line = line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._sep)]
                    temp = [x for x in temp if x]
                    self.__Database.append(temp)
            else:
                try:
                    with open(self._iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line = line.strip()
                            temp = [i.rstrip() for i in line.split(self._sep)]
                            temp = [x for x in temp if x]
                            self.__Database.append(temp)
                except IOError:
                    print("File Not Found")
                    quit()

    def _getMISValues(self):
        """
        Storing the Minimum supports given by the user for each item in the database

        """
        self._MISValues = {}
        if isinstance(self._MIS, _fp._pd.DataFrame):
            items, MIS = [], []
            if self._MIS.empty:
                print("its empty..")
            i = self._MIS.columns.values.tolist()
            if 'items' in i:
                items = self._MIS['items'].tolist()
            if 'MIS' in i:
                MIS = self._MIS['MIS'].tolist()
            for i in range(len(items)):
                self._MISValues[items[i]] = MIS[i]

        if isinstance(self._MIS, str):
            if _fp._validators.url(self._MIS):
                data = _fp._urlopen(self._MIS)
                for line in data:
                    line = line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._sep)]
                    temp = [x for x in temp if x]
                    self._MISValues[temp[0]] = int(temp[1])
            else:
                try:
                    with open(self._MIS, 'r', encoding='utf-8') as f:
                        for line in f:
                            line = line.strip()
                            temp = [i.rstrip() for i in line.split(self._sep)]
                            temp = [x for x in temp if x]
                            self._MISValues[temp[0]] = int(temp[1])
                except IOError:
                    print("File Not Found")
                    quit()

    def __convert(self, value):
        """
        to convert the type of user specified minSup value

        :param value: user specified minSup value
        :return: converted type
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (len(self.__Database) * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (len(self.__Database) * value)
            else:
                value = int(value)
        return value

    def __frequentOneItem(self):
        """
        Generating One frequent items sets
        """
        global minMIS
        self.__mapSupport = {}
        for tr in self.__Database:
            for i in range(0, len(tr)):
                if tr[i] not in self.__mapSupport:
                    self.__mapSupport[tr[i]] = 1
                else:
                    self.__mapSupport[tr[i]] += 1
        self.__mapSupport = {k: v for k, v in self.__mapSupport.items() if v >= min(self._MISValues.values())}
        minMIS = min(self._MISValues.values())
        genList = [k for k, v in sorted(self.__mapSupport.items(), key=lambda x: x[1], reverse=True)]
        self.__rank = dict([(index, item) for (item, index) in enumerate(genList)])
        return genList

    def __updateTransactions(self, itemSet):
        """
        Updates the items in transactions with rank of items according to their support

        :Example: oneLength = {'a':7, 'b': 5, 'c':'4', 'd':3}
                    rank = {'a':0, 'b':1, 'c':2, 'd':3}

        :param itemSet: list of one-frequent items

        """
        list1 = []
        for tr in self.__Database:
            list2 = []
            for i in range(len(tr)):
                if tr[i] in itemSet:
                    list2.append(self.__rank[tr[i]])
            if len(list2) >= 1:
                list2.sort()
                list1.append(list2)
        return list1

    @staticmethod
    def __buildTree(transactions, info):
        """
        Builds the tree with updated transactions

        :param transactions: updated transactions
        :param info: support details of each item in transactions.
        :return: transactions compressed in fp-tree
        """
        rootNode = _Tree()
        rootNode.info = info.copy()
        for i in range(len(transactions)):
            rootNode.addTransaction(transactions[i], 1)
        return rootNode

    def __savePeriodic(self, itemSet):
        """
        The duplication items and their ranks

        :param itemSet: frequent itemSet that generated
        :return: patterns with original item names.

        """
        temp = str()
        for i in itemSet:
            temp = temp + self.__rankDup[i] + " "
        return temp

    @deprecated("It is recommended to use mine() instead of startMine() for mining process")
    def startMine(self):
        """
        main program to start the operation

        """
        self.mine()

    def mine(self):
        """
        main program to start the operation

        """
        global MIS
        self.__startTime = _fp._time.time()
        if self._iFile is None:
            raise Exception("Please enter the file path or file name:")
        self.__creatingItemSets()
        self._getMISValues()
        itemSet = self.__frequentOneItem()
        updatedTransactions = self.__updateTransactions(itemSet)
        for x, y in self.__rank.items():
            MIS[y] = self._MISValues[x]
            self.__rankDup[y] = x
        info = {self.__rank[k]: v for k, v in self.__mapSupport.items()}
        __Tree = self.__buildTree(updatedTransactions, info)
        patterns = __Tree.generatePatterns([])
        self.__finalPatterns = {}
        for k in patterns:
            s = self.__savePeriodic(k[0])
            self.__finalPatterns[str(s)] = k[1]
        print("Frequent patterns were generated successfully using Conditional Frequent Pattern Growth algorithm")
        self.__endTime = _fp._time.time()
        self.__memoryUSS = float()
        self.__memoryRSS = float()
        process = _fp._psutil.Process(_fp._os.getpid())
        self.__memoryUSS = process.memory_full_info().uss
        self.__memoryRSS = process.memory_info().rss


    def getMemoryUSS(self):
        """Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self.__memoryUSS

    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self.__memoryRSS

    def getRuntime(self):
        """Calculating the total amount of runtime taken by the mining process


        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self.__endTime - self.__startTime

    def getPatternsAsDataFrame(self):
        """Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataframe = {}
        data = []
        for a, b in self.__finalPatterns.items():
            data.append([a, b])
            dataframe = _fp._pd.DataFrame(data, columns=['Patterns', 'Support'])
        return dataframe

    def save(self, outFile):
        """Complete set of frequent patterns will be loaded in to a output file

        :param outFile: name of the output file
        :type outFile: file
        """
        self._oFile = outFile
        writer = open(self._oFile, 'w+')
        for x, y in self.__finalPatterns.items():
            s1 = x + ":" + str(y)
            writer.write("%s \n" % s1)

    def getPatterns(self):
        """ Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns
        :rtype: dict
        """
        return self.__finalPatterns

    def printResults(self) -> None:
        """
        this function is used to print the results
        :return: None
        """
        print("Total number of  Frequent Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in ms:",  self.getRuntime())


if __name__ == "__main__":
    _ap = str()
    if len(_fp._sys.argv) == 4 or len(_fp._sys.argv) == 5:
        if len(_fp._sys.argv) == 5:
            _ap = CFPGrowthPlus(_fp._sys.argv[1], _fp._sys.argv[3], _fp._sys.argv[4])
        if len(_fp._sys.argv) == 4:
            _ap = CFPGrowthPlus(_fp._sys.argv[1], _fp._sys.argv[3])
        _ap.startMine()
        _Patterns = _ap.getPatterns()
        print("Total number of Frequent Patterns:", len(_Patterns))
        _ap.savePatterns(_fp._sys.argv[2])
        _memUSS = _ap.getMemoryUSS()
        print("Total Memory in USS:", _memUSS)
        _memRSS = _ap.getMemoryRSS()
        print("Total Memory in RSS", _memRSS)
        _run = _ap.getRuntime()
        print("Total ExecutionTime in ms:", _run)
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")


# findNeighboursUsingGeodesic is a code used to create a neighbourhood file using Geodesic distance.
#
#  **Importing this algorithm into a python program**
# --------------------------------------------------------
#
#     from PAMI.extras.neighbours import findNeighboursUsingGeodesic as db
#
#     obj = db.findNeighboursUsingGeodesic(iFile, oFile, 10, "\t")
#
#     obj.save()
#




__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import sys
import re
from math import sqrt
from geopy.distance import geodesic


class createNeighborhoodFileUsingGeodesicDistance:
    """
    This class create a neighbourhood file using Geodesic distance.

    :Attribute:

        :param iFile : file
            Input file name or path of the input file
        :param oFile : file
            Output file name or path pf the output file
        :param maxDistance : float
            The user can specify maxDistance in Km(Kilometers).
            This program find pairs of values whose Geodesic distance is less than or equal to maxDistace
            and store the pairs.
        :param  seperator: str :
                    This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.


    :Methods:

        mine()
            find and store the pairs of values whose Geodesic distance is less than or equal to maxDistace.
        getFileName()
            This function returns output file name.

    **Importing this algorithm into a python program**
    --------------------------------------------------------
    .. code-block:: python

            from PAMI.extras.neighbours import findNeighboursUsingGeodesic as db

             obj = db.findNeighboursUsingGeodesic(iFile, oFile, 10, "\t")

            obj.save()
    """

    def __init__(self, iFile: str, oFile: str, maxDistance: float, seperator='\t'):
        self.iFile = iFile
        self.oFile = oFile
        self.maxDistance = maxDistance

        coordinates = []
        result = {}
        with open(self.iFile, "r") as f:
            for line in f:
                l = line.rstrip().split(seperator)
                # print(l)
                l[2] = re.sub(r'[^0-9. ]', '', l[2])
                coordinates.append(l[2].rstrip().split(' '))
                # print(l[0])
        for i in range(len(coordinates)):
            for j in range(len(coordinates)):
                if i != j:
                    firstCoordinate = coordinates[i]
                    secondCoordinate = coordinates[j]
                    long1 = float(firstCoordinate[0])
                    lat1 = float(firstCoordinate[1])
                    long2 = float(secondCoordinate[0])
                    lat2 = float(secondCoordinate[1])

                    dist = geodesic((lat1, long1), (lat2, long2)).kilometers

                    if dist <= float(self.maxDistance):
                        result[tuple(firstCoordinate)] = result.get(tuple(firstCoordinate), [])
                        result[tuple(firstCoordinate)].append(secondCoordinate)

        with open(self.oFile, "w+") as f:
            for i in result:
                string = "Point(" + i[0] + " " + i[1] + ")" + seperator
                f.write(string)
                for j in result[i]:
                    string = "Point(" + j[0] + " " + j[1] + ")" + seperator
                    f.write(string)
                f.write("\n")

    def getFileName(self):
        return self.oFile


if __name__ == "__main__":
    obj = createNeighborhoodFileUsingGeodesicDistance(sys.argv[1], sys.argv[2], sys.argv[4])



import random
def generate_random_graph(n, m, max_label,f):
    vertices = [(f"v {i}", random.randint(1, max_label)) for i in range(n)]
    edges = set()

    while len(edges) < m:
        u = random.randint(0, n - 1)
        v = random.randint(0, n - 1)
        label = random.randint(1, max_label)

        if u != v and (u, v, label) not in edges:
            edges.add((u, v, label))

    for vertex in vertices:
        s=f"{vertex[0]} {vertex[1]}\n"
        f.write(s)

    for edge in edges:
        s=f"e {edge[0]} {edge[1]} {edge[2]}\n"
        f.write(s)


def generate(path, transactions):
    f=open(path,"w+")
    for t in range(transactions):
        f.write("t # {}\n".format(t))
        generate_random_graph(random.randint(10,20),random.randint(20,30),3,f)

    f.write("t # -1")
    f.close()



# Spatial High Utility ItemSet Mining (SHUIM) [3] is an important model in data
# mining with many real-world applications. It involves finding all spatially interesting itemSets having high value
# in a quantitative spatio-temporal database.
#
# **Importing this algorithm into a python program**
# --------------------------------------------------------
#
#
#             from PAMI.highUtilitySpatialPattern.basic import HDSHUIM as alg
#
#             obj=alg.HDSHUIM("input.txt","Neighbours.txt",35)
#
#             obj.mine()
#
#             Patterns = obj.getPatterns()
#
#             print("Total number of Spatial High-Utility Patterns:", len(Patterns))
#
#             obj.save("output")
#
#             memUSS = obj.getMemoryUSS()
#
#             print("Total Memory in USS:", memUSS)
#
#             memRSS = obj.getMemoryRSS()
#
#             print("Total Memory in RSS", memRSS)
#
#             run = obj.getRuntime()
#
#             print("Total ExecutionTime in seconds:", run)
#





__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
     Copyright (C)  2021 Rage Uday Kiran

"""

from PAMI.highUtilitySpatialPattern.basic import abstract as _ab
from typing import List, Dict, Tuple, Set, Union, Any, Generator
from deprecated import deprecated

class _Element:
    """
    A class represents an Element of a utility list as used by the HDSHUIM algorithm.

    :Attributes:

        ts : int
            keep tact of transaction id
        snu : int
            Spatial non-closed itemSet utility
        remainingUtility : int
            Spatial non-closed remaining utility
        pu : int
            prefix utility
        prevPos: int
            position of previous item in the list
    """

    def __init__(self, ts: int, snu: int, remainingUtility: int, pu: int, prevPos: int) -> None:
        self.ts = ts
        self.snu = snu
        self.remainingUtility = remainingUtility
        self.pu = pu
        self.prevPos = prevPos


class _CUList:
    """
        A class represents a UtilityList as used by the HDSHUIM algorithm.

    :Attributes:

        item: int
            item 
        sumSnu: long
            the sum of item utilities
        sumRemainingUtility: long
            the sum of remaining utilities
        sumCu : long
            the sum of closed utilities
        sumCru: long
            the sum of closed remaining utilities
        sumCpu: long
            the sum of closed prefix utilities
        elements: list
            the list of elements 

    :Methods:

        addElement(element)
            Method to add an element to this utility list and update the sums at the same time.

    """

    def __init__(self, item: str) -> None:
        self.item = item
        self.sumSnu = 0
        self.sumRemainingUtility = 0
        self.sumCu = 0
        self.sumCru = 0
        self.sumCpu = 0
        self.elements = []

    def addElements(self, element: _Element) -> None:
        """
        A method to add new element to CUList

        :param element: element to be added to CUList
        :type element: Element
        :return: None
        """
        self.sumSnu += element.snu
        self.sumRemainingUtility += element.remainingUtility
        self.elements.append(element)


class _Pair:
    """
    A class represent an item and its utility in a transaction
    """

    def __init__(self) -> None:
        self.item = 0
        self.utility = 0


class HDSHUIM(_ab._utilityPatterns):
    """
    :Description:

        Spatial High Utility ItemSet Mining (SHUIM) [3] is an important model in data
        mining with many real-world applications. It involves finding all spatially interesting itemSets having high value 
        in a quantitative spatio temporal database.

    :Reference:

        P. Pallikila et al., "Discovering Top-k Spatial High Utility Itemsets in Very Large Quantitative Spatiotemporal 
        databases," 2021 IEEE International Conference on Big Data (Big Data), Orlando, FL, USA, 2021, pp. 4925-4935,
        doi: 10.1109/BigData52589.2021.9671912.

    :param  iFile: str :
                   Name of the Input file to mine complete set of High Utility Spatial patterns
    :param  oFile: str :
                   Name of the output file to store complete set of High Utility Spatial patterns
    :param  minSup: int or float or str :
                   The user can specify minSup either in count or proportion of database size. If the program detects the data type of minSup is integer, then it treats minSup is expressed in count. Otherwise, it will be treated as float.
    :param maxPer: float :
                   The user can specify maxPer in count or proportion of database size. If the program detects the data type of maxPer is integer, then it treats maxPer is expressed in count.
    :param minUtil: int :
                   Minimum utility threshold given by User
    :param nFile: str :
                   Name of the input file to mine complete set of High Utility Spatial patterns
    :param  sep: str :
                   This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.


    :Attributes:

        iFile : str
            Name of the input file to mine complete set of frequent patterns
        oFile : str
            Name of the output file to store complete set of frequent patterns
        nFile: str
           Name of Neighbourhood items file
        memoryRSS : float
            To store the total amount of RSS memory consumed by the program
        startTime:float
            To record the start time of the mining process
        endTime:float
            To record the completion time of the mining process
        minUtil : int
            The user given minUtil
        mapFMAP: list
            EUCS map of the FHM algorithm
        candidates: int
            candidates generated
        huiCnt: int
            huis created
        neighbors: map
            keep track of neighbours of elements
        mapOfPMU: map
            a map to keep track of Probable Maximum utility(PMU) of each item
    :Methods:

            mine()
                Mining process will start from here
            getPatterns()
                Complete set of patterns will be retrieved with this function
            save(oFile)
                Complete set of frequent patterns will be loaded in to a output file
            constructCUL(x, compactUList, st, minUtil, length, exNeighbours)
                A method to construct CUL's database
            getPatternsAsDataFrame()
                Complete set of frequent patterns will be loaded in to a dataframe
            getMemoryUSS()
                Total amount of USS memory consumed by the mining process will be retrieved from this function
            getMemoryRSS()
                Total amount of RSS memory consumed by the mining process will be retrieved from this function
            getRuntime()
                Total amount of runtime taken by the mining process will be retrieved from this function
            Explore_SearchTree(prefix, uList, exNeighbours, minUtil)
                A method to find all high utility itemSets
            updateClosed(x, compactUList, st, exCul, newT, ex, eyTs, length)
                A method to update closed values
            saveItemSet(prefix, prefixLen, item, utility)
               A method to save itemSets
            updateElement(z, compactUList, st, exCul, newT, ex, duPrevPos, eyTs)
               A method to updates vales for duplicates


    **Executing the code on terminal:**
    -------------------------------------

    .. code-block:: console

      Format:

      (.venv) $ python3 HDSHUIM.py <inputFile> <outputFile> <Neighbours> <minUtil> <separator>

      Example Usage:

      (.venv) $ python3 HDSHUIM.py sampleTDB.txt output.txt sampleN.txt 35 ','

    .. note:: minSup will be considered in percentage of database transactions


    **Sample run of importing the code:**
    ---------------------------------------
    .. code-block:: python
        
            from PAMI.highUtilityGeoreferencedFrequentPattern.basic import HDSHUIM as alg

            obj=alg.HDSHUIM("input.txt","Neighbours.txt",35)

            obj.mine()

            Patterns = obj.getPatterns()

            print("Total number of Spatial High-Utility Patterns:", len(Patterns))

            obj.save("output")

            memUSS = obj.getMemoryUSS()

            print("Total Memory in USS:", memUSS)

            memRSS = obj.getMemoryRSS()

            print("Total Memory in RSS", memRSS)

            run = obj.getRuntime()

            print("Total ExecutionTime in seconds:", run)

    **Credits:**
    ---------------
        The complete program was written by B.Sai Chitra under the supervision of Professor Rage Uday Kiran.
            
    """

    _startTime = float()
    _endTime = float()
    _minSup = str()
    _maxPer = float()
    _finalPatterns = {}
    _iFile = " "
    _oFile = " "
    _nFile = " "
    _minUtil = 0
    _memoryUSS = float()
    _memoryRSS = float()
    _sep = "\t"

    def __init__(self, iFile: str, nFile: str, minUtil: int, sep: str="\t") -> None:
        super().__init__(iFile, nFile, minUtil, sep)
        self._startTime = 0
        self._endTime = 0
        self._huiCount = 0
        self._candidates = 0
        self._mapOfPMU = {}
        self._mapFMAP = {}
        self._neighbors = {}
        self._finalPatterns = {}

    def _compareItems(self, o1: Any, o2: Any) -> int:
        """
        A Function that sort all FFI-list in ascending order of Support

        :param o1: First FFI-list

        :type o1: _FFList

        :param o2: Second FFI-list

        :type o1: _FFList

        :return: Comparision Value

        :rtype: int
        """
        compare = self._mapOfPMU[o1.item] - self._mapOfPMU[o2.item]
        if compare == 0:
            return int(o1.item) - int(o2.item)
        else:
            return compare

    @deprecated("It is recommended to use 'mine()' instead of 'startMine()' for mining process. Starting from January 2025, 'startMine()' will be completely terminated.")
    def startMine(self) -> None:
        """
        main program to start the operation
        """
        self.mine()



    def mine(self) -> None:
        """
        main program to start the operation
        """
        minUtil = self._minUtil
        self._startTime = _ab._time.time()
        with open(self._nFile, 'r') as file1:
            for line in file1:
                line = line.split("\n")[0]
                parts = line.split(self._sep)
                parts = [i.strip() for i in parts]
                item = parts[0]
                neigh1 = list()
                for i in range(1, len(parts)):
                    neigh1.append(parts[i])
                self._neighbors[item] = set(neigh1)
        with open(self._iFile, 'r') as file:
            for line in file:
                parts = line.split(":")
                itemString = (parts[0].split("\n")[0]).split(self._sep)
                utilityString = (parts[2].split("\n")[0]).split(self._sep)
                transUtility = int(parts[1])
                trans1 = set()
                for i in range(0, len(itemString)):
                    trans1.add(itemString[i])
                for i in range(0, len(itemString)):
                    item = itemString[i]
                    twu = self._mapOfPMU.get(item)
                    if twu is None:
                        twu = int(utilityString[i])
                    else:
                        twu += int(utilityString[i])
                    self._mapOfPMU[item] = twu
                    if self._neighbors.get(item) is None:
                        continue
                    neighbours2 = trans1.intersection(self._neighbors.get(item))
                    for item2 in neighbours2:
                        if self._mapOfPMU.get(item2) is None:
                            self._mapOfPMU[item2] = int(utilityString[i])
                        else:
                            self._mapOfPMU[item2] += int(utilityString[i])

        listOfCUList = []
        hashTable = {}
        mapItemsToCUList = {}
        for item in self._mapOfPMU.keys():
            if self._mapOfPMU.get(item) >= minUtil:
                uList = _CUList(item)
                mapItemsToCUList[item] = uList
                listOfCUList.append(uList)
        listOfCUList.sort(key=_ab._functools.cmp_to_key(self._compareItems))
        ts = 1
        with open(self._iFile, 'r') as file:
            for line in file:
                parts = line.split(":")
                items = (parts[0].split("\n")[0]).split(self._sep)
                utilities = (parts[2].split("\n")[0]).split(self._sep)
                ru = 0
                newTwu = 0
                txKey = []
                revisedTrans = []
                for i in range(0, len(items)):
                    pair = _Pair()
                    pair.item = items[i]
                    pair.utility = int(utilities[i])
                    if self._mapOfPMU.get(pair.item) >= minUtil:
                        revisedTrans.append(pair)
                        txKey.append(pair.item)
                        newTwu += pair.utility
                revisedTrans.sort(key=_ab._functools.cmp_to_key(self._compareItems))
                txKey1 = tuple(txKey)
                if len(revisedTrans) > 0:
                    if txKey1 not in hashTable.keys():
                        hashTable[txKey1] = len(mapItemsToCUList[revisedTrans[len(revisedTrans) - 1].item].elements)
                        for i in range(len(revisedTrans) - 1, -1, -1):
                            pair = revisedTrans[i]
                            cuListOfItems = mapItemsToCUList.get(pair.item)
                            element = _Element(ts, pair.utility, ru, 0, 0)
                            if i > 0:
                                element.prevPos = len(mapItemsToCUList[revisedTrans[i - 1].item].elements)
                            else:
                                element.prevPos = -1
                            cuListOfItems.addElements(element)
                            ru += pair.utility
                    else:
                        pos = hashTable[txKey1]
                        ru = 0
                        for i in range(len(revisedTrans) - 1, -1, -1):
                            cuListOfItems = mapItemsToCUList[revisedTrans[i].item]
                            cuListOfItems.elements[pos].snu += revisedTrans[i].utility
                            cuListOfItems.elements[pos].remainingUtility += ru
                            cuListOfItems.sumSnu += revisedTrans[i].utility
                            cuListOfItems.sumRemainingUtility += ru
                            ru += revisedTrans[i].utility
                            pos = cuListOfItems.elements[pos].prevPos
                # EUCS
                for i in range(len(revisedTrans) - 1, -1, -1):
                    pair = revisedTrans[i]
                    mapFMAPItem = self._mapFMAP.get(pair.item)
                    if mapFMAPItem is None:
                        mapFMAPItem = {}
                        self._mapFMAP[pair.item] = mapFMAPItem
                    for j in range(i + 1, len(revisedTrans)):
                        pairAfter = revisedTrans[j]
                        twuSUm = mapFMAPItem.get(pairAfter.item)
                        if twuSUm is None:
                            mapFMAPItem[pairAfter.item] = newTwu
                        else:
                            mapFMAPItem[pairAfter.item] = twuSUm + newTwu
                ts += 1
        exNeighbours = set(self._mapOfPMU.keys())
        # print(self.Neighbours)
        self._ExploreSearchTree([], listOfCUList, exNeighbours, minUtil)
        self._endTime = _ab._time.time()
        process = _ab._psutil.Process(_ab._os.getpid())
        self._memoryUSS = float()
        self._memoryRSS = float()
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss

    def _ExploreSearchTree(self, prefix: List[str], uList: List[_CUList], exNeighbours: set, minUtil: int) -> None:
        """
        A method to find all high utility itemSets
        :parm prefix: it represents all items in prefix
        :type prefix :list
        :parm uList:projected Utility list.
        :type uList: list
        :parm exNeighbours: keep track of common Neighbours
        :type exNeighbours: set
        :parm minUtil:user minUtil
        :type minUtil:int
        :return: None
        """
        for i in range(0, len(uList)):
            x = uList[i]
            if x.item not in exNeighbours:
                continue
            self._candidates += 1
            sortedPrefix = [0] * (len(prefix) + 1)
            sortedPrefix = prefix[0:len(prefix) + 1]
            sortedPrefix.append(x.item)
            if (x.sumSnu + x.sumCu >= minUtil) and (x.item in exNeighbours):
                self._saveItemSet(prefix, len(prefix), x.item, x.sumSnu + x.sumCu)
            if x.sumSnu + x.sumCu + x.sumRemainingUtility + x.sumCru >= minUtil:  # U-Prune # and (x.item in exNeighbours)):
                ULIST = []
                for j in range(i, len(uList)):
                    if (uList[j].item in exNeighbours) and (self._neighbors.get(x.item) is not None) and (
                            uList[j].item in self._neighbors.get(x.item)):
                        ULIST.append(uList[j])
                exULs = self._constructCUL(x, ULIST, -1, minUtil, len(sortedPrefix), exNeighbours)
                if self._neighbors.get(x.item) is not None and exNeighbours is not None:
                    set1 = exNeighbours.intersection(self._neighbors.get(x.item))
                    if exULs is None or set1 is None:
                        continue
                    self._ExploreSearchTree(sortedPrefix, exULs, set1, minUtil)

    def _constructCUL(self, x: _Element, compactUList: List[_CUList], st: int, minUtil: int, length: int, exNeighbours: set) -> List[_CUList]:
        """
        A method to construct CUL's database
        :parm x: Compact utility list
        :type x: Node
        :parm compactUList:list of Compact utility lists.
        :type compactUList:list
        :parm st: starting pos of compactUList
        :type st:int
        :parm minUtil: user minUtil
        :type minUtil:int
        :parm length: length of x
        :type length:int
        :parm exNeighbours: common Neighbours
        :type exNeighbours: set
        :return: projected database of list X
        :rtype: list or set
        """
        exCul = []
        lau = []
        cUtil = []
        eyTs = []
        for i in range(0, len(compactUList)):
            uList = _CUList(compactUList[i].item)
            exCul.append(uList)
            lau.append(0)
            cUtil.append(0)
            eyTs.append(0)
        sz = len(compactUList) - (st + 1)
        exSZ = sz
        for j in range(st + 1, len(compactUList)):
            mapOfTWUF = self._mapFMAP[x.item]
            if mapOfTWUF is not None:
                twuf = mapOfTWUF.get(compactUList[j].item)
                if twuf != None and twuf < minUtil or (not (exCul[j].item in exNeighbours)):
                    exCul[j] = None
                    exSZ = sz - 1
                else:
                    uList = _CUList(compactUList[j].item)
                    exCul[j] = uList
                    eyTs[j] = 0
                    lau[j] = x.sumCu + x.sumCru + x.sumSnu + x.sumRemainingUtility
                    cUtil[j] = x.sumCu + x.sumCru
        hashTable = {}
        for ex in x.elements:
            newT = []
            for j in range(st + 1, len(compactUList)):
                if exCul[j] is None:
                    continue
                eyList = compactUList[j].elements
                while eyTs[j] < len(eyList) and eyList[eyTs[j]].ts < ex.ts:
                    eyTs[j] = eyTs[j] + 1
                if eyTs[j] < len(eyList) and eyList[eyTs[j]].ts == ex.ts:
                    newT.append(j)
                else:
                    lau[j] = lau[j] - ex.snu - ex.remainingUtility
                    if lau[j] < minUtil:
                        exCul[j] = None
                        exSZ = exSZ - 1
            if len(newT) == exSZ:
                self._updateClosed(x, compactUList, st, exCul, newT, ex, eyTs, length)
            else:
                if len(newT) == 0:
                    continue
                ru = 0
                newT1 = tuple(newT)
                if newT1 not in hashTable.keys():
                    hashTable[newT1] = len(exCul[newT[len(newT) - 1]].elements)
                    for i in range(len(newT) - 1, -1, -1):
                        cuListOfItems = exCul[newT[i]]
                        y = compactUList[newT[i]].elements[eyTs[newT[i]]]
                        element = _Element(ex.ts, ex.snu + y.snu - ex.pu, ru, ex.snu, 0)
                        if i > 0:
                            element.prevPos = len(exCul[newT[i - 1]].elements)
                        else:
                            element.prevPos = -1
                        cuListOfItems.addElements(element)
                        ru += y.snu - ex.pu
                else:
                    dPrevPos = hashTable[newT1]
                    self._updateElement(x, compactUList, st, exCul, newT, ex, dPrevPos, eyTs)
            for j in range(st + 1, len(compactUList)):
                cUtil[j] = cUtil[j] + ex.snu + ex.remainingUtility
        filter_compactUList = []
        for j in range(st + 1, len(compactUList)):
            if cUtil[j] < minUtil or exCul[j] is None:
                continue
            else:
                if length > 1:
                    exCul[j].sumCu += compactUList[j].sumCu + x.sumCu - x.sumCpu
                    exCul[j].sumCru += compactUList[j].sumCru
                    exCul[j].sumCpu += x.sumCu
                filter_compactUList.append(exCul[j])
        return filter_compactUList

    def _updateClosed(self, x: _Element, compactUList: List[_CUList], st: int, exCul: List[_CUList], newT: List[int], ex: _Element, eyTs: List[int], length: int) -> None:
        """
        A method to update closed values
        :parm x: Compact utility list.
        :type x: list
        :parm compactUList:list of Compact utility lists.
        :type compactUList:list
        :parm st: starting pos of compactUList
        :type st:int
        :parm newT:transaction to be updated
        :type newT:list
        :parm ex: element ex
        :type ex:element
        :parm eyTs:list of tss
        :type eyTs:ts
        :parm length: length of x
        :type length:int
        :return: None
        """
        remainingUtility = 0
        for j in range(len(newT) - 1, -1, -1):
            ey = compactUList[newT[j]]
            eyy = ey.elements[eyTs[newT[j]]]
            exCul[newT[j]].sumCu += ex.snu + eyy.snu - ex.pu
            exCul[newT[j]].sumCru += remainingUtility
            exCul[newT[j]].sumCpu += ex.snu
            remainingUtility = remainingUtility + eyy.snu - ex.pu

    def _updateElement(self, z: _Element, compactUList: List[_CUList], st: int, exCul: List[_CUList], newT: List[int], ex: _Element, duPrevPos: int, eyTs: List[int]) -> None:
        """
        A method to updates vales for duplicates
        :parm z: Compact utility list
        :type z: list
        :parm compactUList:list of Compact utility lists
        :type compactUList:list
        :parm st: starting pos of compactUList
        :type st:int
        :parm exCul:list of compactUList
        :type exCul:list
        :parm newT:transaction to be updated
        :type newT:list
        :parm ex: element ex
        :type ex:element
        :parm duPrevPos: position of z in exCul
        :type duPrevPos:int
        :parm eyTs:list of tss
        :type eyTs:ts
        :return: None
        """
        remainingUtility = 0
        pos = duPrevPos
        for j in range(len(newT) - 1, -1, -1):
            ey = compactUList[newT[j]]
            eyy = ey.elements[eyTs[newT[j]]]
            exCul[newT[j]].elements[pos].snu += ex.snu + eyy.snu - ex.pu
            exCul[newT[j]].sumSnu += ex.snu + eyy.snu - ex.pu
            exCul[newT[j]].elements[pos].remainingUtility += remainingUtility
            exCul[newT[j]].sumRemainingUtility += remainingUtility
            exCul[newT[j]].elements[pos].pu += ex.snu
            remainingUtility = remainingUtility + eyy.snu - ex.pu
            pos = exCul[newT[j]].elements[pos].prevPos

    def _saveItemSet(self, prefix: List[str], prefixLen: int, item: str, utility: int) -> None:
        """
         A method to save itemSets
        :parm prefix: it represents all items in prefix
        :type prefix :list
        :parm item:item
        :type item: int
        :parm utility:utility of itemSet
        :type utility:int
        :return: None
        """
        self._huiCount += 1
        res = str()
        for i in range(0, prefixLen):
            res += str(prefix[i]) + "\t"
        res += str(item)
        res1 = str(utility)
        self._finalPatterns[res] = res1

    def getPatternsAsDataFrame(self) -> Dict[str, str]:
        """
        Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataFrame = {}
        data = []
        for a, b in self._finalPatterns.items():
            data.append([a.replace('\t', ' '), b])
            dataFrame = _ab._pd.DataFrame(data, columns=['Patterns', 'Support'])
        return dataFrame

    def getPatterns(self) -> Dict[str, str]:
        """
        Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns
        :rtype: dict
        """
        return self._finalPatterns

    def save(self, outFile: str) -> None:
        """
        Complete set of frequent patterns will be loaded in to an output file

        :param outFile: name of the output file
        :type outFile: csv file
        :return: None
        """
        self.oFile = outFile
        writer = open(self.oFile, 'w+')
        for x, y in self._finalPatterns.items():
            patternsAndSupport = x.strip() + ":" + str(y)
            writer.write("%s \n" % patternsAndSupport)

    def getMemoryUSS(self) -> float:
        """
        Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self) -> float:
        """
        Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """
        return self._memoryRSS

    def getRuntime(self) -> float:
        """
        Calculating the total amount of runtime taken by the mining process

        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """
        return self._endTime - self._startTime

    def printResults(self) -> None:
        """
        This function is used to print the results
        """
        print("Total number of Spatial High Utility Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in seconds:", self.getRuntime())


if __name__ == "__main__":
    _ap = str()
    if len(_ab._sys.argv) == 5 or len(_ab._sys.argv) == 6:
        if len(_ab._sys.argv) == 6:  # to  include a user specified separator
            _ap = HDSHUIM(_ab._sys.argv[1], _ab._sys.argv[3], int(_ab._sys.argv[4]), _ab._sys.argv[5])
        if len(_ab._sys.argv) == 5:  # to consider "\t" as a separator
            _ap = HDSHUIM(_ab._sys.argv[1], _ab._sys.argv[3], int(_ab._sys.argv[4]))
        _ap.startMine()
        _ap.mine()
        print("Total number of Spatial High-Utility Patterns:", len(_ap.getPatterns()))
        _ap.save(_ab._sys.argv[2])
        print("Total Memory in USS:", _ap.getMemoryUSS())
        print("Total Memory in RSS", _ap.getMemoryRSS())
        print("Total ExecutionTime in ms:", _ap.getRuntime())
    else:
        for i in [100000, 500000]:
            _ap = HDSHUIM('/Users/Likhitha/Downloads/mushroom_main_2000.txt',
                    '/Users/Likhitha/Downloads/mushroom_neighbors_2000.txt', i, ' ')
            _ap.startMine()
            _ap.mine()
            print("Total number of Spatial High Utility Patterns:", len(_ap.getPatterns()))
            print("Total Memory in USS:", _ap.getMemoryUSS())
            print("Total Memory in RSS", _ap.getMemoryRSS())
            print("Total ExecutionTime in seconds:", _ap.getRuntime())
        print("Error! The number of input parameters do not match the total number of parameters provided")



# 3pEclat is the fundamental approach to mine the partial periodic frequent patterns.
#
# **Importing this algorithm into a python program**
# --------------------------------------------------------
#
#             from PAMI.periodicFrequentPattern.basic import PPP_ECLAT as alg
#
#             obj = alg.PPP_ECLAT(iFile, minPS, period)
#
#             obj.startMine()
#
#             Patterns = obj.getPatterns()
#
#             print("Total number of partial periodic patterns:", len(Patterns))
#
#             obj.save(oFile)
#
#             Df = obj.getPatternsAsDataFrame()
#
#             memUSS = obj.getMemoryUSS()
#
#             print("Total Memory in USS:", memUSS)
#
#             memRSS = obj.getMemoryRSS()
#
#             print("Total Memory in RSS", memRSS)
#
#             run = obj.getRuntime()
#
#             print("Total ExecutionTime in seconds:", run)
#


__copyright__ = """
 Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
     Copyright (C)  2021 Rage Uday Kiran

"""



from PAMI.partialPeriodicPattern.basic import abstract as _ab
from typing import List, Dict, Tuple, Set, Union, Any, Generator
import pandas as pd
import numpy as np
from deprecated import deprecated

class PPP_ECLAT(_ab._partialPeriodicPatterns):
    """
    :Descripition:   3pEclat is the fundamental approach to mine the partial periodic frequent patterns.

    :Reference:   R. Uday Kirana,b,∗ , J.N. Venkateshd, Masashi Toyodaa , Masaru Kitsuregawaa,c , P. Krishna Reddy Discovering partial periodic-frequent patterns in a transactional database
                  https://www.tkl.iis.u-tokyo.ac.jp/new/uploads/publication_file/file/774/JSS_2017.pdf

    :param  iFile: str :
                   Name of the Input file to mine complete set of frequent pattern's
    :param  oFile: str :
                   Name of the output file to store complete set of frequent patterns
    :param  minPS: float:
                   Minimum partial periodic pattern...
    :param  period: float:
                   Minimum partial periodic...

    :param  sep: str :
                   This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.

    :Attributes:

        self.iFile : file
            Name of the Input file or path of the input file
        self. oFile : file
            Name of the output file or path of the output file
        minPS: float or int or str
            The user can specify minPS either in count or proportion of database size.
            If the program detects the data type of minPS is integer, then it treats minPS is expressed in count.
            Otherwise, it will be treated as float.
            Example: minPS=10 will be treated as integer, while minPS=10.0 will be treated as float
        period: float or int or str
            The user can specify period either in count or proportion of database size.
            If the program detects the data type of period is integer, then it treats period is expressed in count.
            Otherwise, it will be treated as float.
            Example: period=10 will be treated as integer, while period=10.0 will be treated as float
        sep : str
            This variable is used to distinguish items from one another in a transaction. The default seperator is tab space or \t.
            However, the users can override their default separator.
        memoryUSS : float
            To store the total amount of USS memory consumed by the program
        memoryRSS : float
            To store the total amount of RSS memory consumed by the program
        startTime:float
            To record the start time of the mining process
        endTime:float
            To record the completion time of the mining process
        Database : list
            To store the transactions of a database in list
        mapSupport : Dictionary
            To maintain the information of item and their frequency
        lno : int
            it represents the total no of transactions
        tree : class
            it represents the Tree class
        finalPatterns : dict
            it represents to store the patterns
        tidList : dict
            stores the timestamps of an item

    :Methods:

        startMine()
            Mining process will start from here
        getPatterns()
            Complete set of patterns will be retrieved with this function
        save(oFile)
            Complete set of frequent patterns will be loaded in to an  output file
        getPatternsAsDataFrame()
            Complete set of frequent patterns will be loaded in to a dataframe
        getMemoryUSS()
            Total amount of USS memory consumed by the mining process will be retrieved from this function
        getMemoryRSS()
            Total amount of RSS memory consumed by the mining process will be retrieved from this function
        getRuntime()
            Total amount of runtime taken by the mining process will be retrieved from this function
        creatingOneitemSets()
            Scan the database and store the items with their timestamps which are periodic frequent
        getPeriodAndSupport()
            Calculates the support and period for a list of timestamps.
        Generation()
            Used to implement prefix class equivalence method to generate the periodic patterns recursively

    **Executing the code on terminal:**
    ----------------------------------------
      .. code-block:: console


       Format:

       (.venv) $ python3 PPP_ECLAT.py <inputFile> <outputFile> <minPS> <period>

       Examples:

       (.venv) $ python3 PPP_ECLAT.py sampleDB.txt patterns.txt 0.3 0.4


    **Sample run of importing the code:**
    -----------------------------------------
    ...     code-block:: python

            from PAMI.periodicFrequentPattern.basic import PPP_ECLAT as alg

            obj = alg.PPP_ECLAT(iFile, minPS,period)

            obj.startMine()

            Patterns = obj.getPatterns()

            print("Total number of partial periodic patterns:", len(Patterns))

            obj.save(oFile)

            Df = obj.getPatternsAsDataFrame()

            memUSS = obj.getMemoryUSS()

            print("Total Memory in USS:", memUSS)

            memRSS = obj.getMemoryRSS()

            print("Total Memory in RSS", memRSS)

            run = obj.getRuntime()

            print("Total ExecutionTime in seconds:", run)

    **Credits:**
    ------------------
    The complete program was written by P.RaviKumar  under the supervision of Professor Rage Uday Kiran.\n

    """

    _startTime = float()
    _endTime = float()
    _finalPatterns = {}
    _iFile = " "
    _oFile = " "
    _sep = " "
    _memoryUSS = float()
    _memoryRSS = float()
    _mapSupport = {}
    _itemsetCount = 0
    _writer = None
    _minPS = str()
    _period = str()
    _tidList = {}
    _lno = 0
    _Database = []

    def _convert(self, value) -> Union[int, float]:
        """
        To convert the given user specified value
        :param value: user specified value

        :return: converted value
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (len(self._Database) * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (len(self._Database) * value)
            else:
                value = int(value)
        return value

    def _getPeriodicSupport(self, timeStamps: list) -> int:
        """
        calculates the support and periodicity with list of timestamps.

        :param timeStamps : timestamps of a pattern
        :type timeStamps : list
        :return: list
        """
        timeStamps.sort()
        per = 0
        for i in range(len(timeStamps) - 1):
            j = i + 1
            if abs(timeStamps[j] - timeStamps[i]) <= self._period:
                per += 1
        return per
    
    def _getPerSup(self, arr):
        arr = list(arr)
        arr.append(self._maxTS)
        arr.append(0)
        arr = np.sort(arr)
        arr = np.diff(arr)

        locs = len(np.where(arr <= self._period)[0])

        return locs

    def _creatingItemSets(self) -> None:
        """
        Storing the complete transactions of the database/input file in a database variable
        :return: None
        """
        self._Database = []
        if isinstance(self._iFile, pd.DataFrame):
            data, ts = [], []
            if self._iFile.empty:
                print("its empty..")
            i = self._iFile.columns.values.tolist()
            if 'TS' in i:
                ts = self._iFile['TS'].tolist()
            if 'Transactions' in i:
                data = self._iFile['Transactions'].tolist()
            for i in range(len(data)):
                if data[i]:
                    tr = [str(ts[i])] + [x for x in data[i].split(self._sep)]
                    self._Database.append(tr)
                else:
                    self._Database.append([str(ts[i])])

        if isinstance(self._iFile, str):
            if _ab._validators.url(self._iFile):
                data = _ab._urlopen(self._iFile)
                for line in data:
                    line.strip()
                    self._lno += 1
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._sep)]
                    temp = [x for x in temp if x]
                    self._Database.append(temp)
            else:
                try:
                    with open(self._iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            self._lno += 1
                            line.strip()
                            temp = [i.rstrip() for i in line.split(self._sep)]
                            temp = [x for x in temp if x]
                            self._Database.append(temp)
                except IOError:
                    print("File Not Found")
                    quit()

    @deprecated("It is recommended to use mine() instead of startMine() for mining process")
    def startMine(self) -> None:
        """
        Main program start with extracting the periodic frequent items from the database and
        performs prefix equivalence to form the combinations and generates partial-periodic patterns.
        :return: None

        """
        self.mine()

    def _getPerSup(self, arr):
        arr = list(arr)
        arr = np.sort(arr)
        arr = np.diff(arr)
        locs = len(np.where(arr <= self._period)[0])

        return locs
    
    def _recursive(self, cands, items):
        for i in range(len(cands)):
            newCands = []
            nitems = {}
            for j in range(i + 1, len(cands)):
                intersection = items[cands[i]].intersection(items[cands[j]])
                perSup = self._getPerSup(intersection)
                if perSup >= self._minPS:
                    nCand = cands[i] + tuple([cands[j][-1]])
                    newCands.append(nCand)
                    nitems[nCand] = intersection
                    self._finalPatterns[nCand] = perSup
                # if len(intersection) >= self._min:
                #     perSup = self._getPerSup(intersection)
                #     ratio = perSup / (len(intersection) + 1)
                #     if ratio >= self._partialPeriodicPatterns__minPR:
                #         nCand = cands[i] + tuple([cands[j][-1]])
                #         newCands.append(nCand)
                #         nitems[nCand] = intersection
                #         self._finalPatterns[nCand] = [len(intersection), ratio]
            if len(newCands) > 1:
                self._recursive(newCands, nitems)


    def mine(self) -> None:
        """
        Main program start with extracting the periodic frequent items from the database and
        performs prefix equivalence to form the combinations and generates partial-periodic patterns.
        :return: None

        """
        self._startTime = _ab._time.time()
        self._creatingItemSets()
        self._finalPatterns = {}


        items = {}
        maxTS = 0
        for line in self._Database:
            index = int(line[0])
            maxTS = max(maxTS, index)
            for item in line[1:]:
                if tuple([item]) not in items:
                    items[tuple([item])] = set()
                items[tuple([item])].add(index)

        self._dbSize = maxTS

        self._period = self._convert(self._period)
        self._minPS = self._convert(self._minPS)

        cands = []
        nitems = {}

        for k, v in items.items():
            perSup = self._getPerSup(v)
            if perSup >= self._minPS:
                self._finalPatterns[k] = perSup
                cands.append(k)
                nitems[k] = v

        self._recursive(cands, nitems)

        temp = {}
        for k,v in self._finalPatterns.items():
            k = list(k)
            k = "\t".join(k)
            temp[k] = v
        self._finalPatterns = temp

        print("Partial Periodic Patterns were generated successfully using 3PEclat algorithm")
        self._endTime = _ab._time.time()
        process = _ab._psutil.Process(_ab._os.getpid())
        self._memoryRSS = float()
        self._memoryUSS = float()
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss

    def getMemoryUSS(self) -> float:
        """
        Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self) -> float:
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryRSS

    def getRuntime(self) -> float:
        """
        Calculating the total amount of runtime taken by the mining process

        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self._endTime - self._startTime

    def getPatternsAsDataFrame(self) -> _ab._pd.DataFrame:
        """Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataframe = {}
        data = []
        for a, b in self._finalPatterns.items():
            data.append([a, b])
        dataframe = _ab._pd.DataFrame(data, columns=['Patterns', 'periodicSupport'])
        return dataframe

    def save(self, outFile: str) -> None:
        """Complete set of frequent patterns will be loaded in to an output file

        :param outFile: name of the output file
        :type outFile: file
        :return: None
        """
        self._oFile = outFile
        writer = open(self._oFile, 'w+')
        for x, y in self._finalPatterns.items():
            s1 = x.strip() + ":" + str(y)
            writer.write("%s \n" % s1)

    def getPatterns(self) -> Dict[str, int]:
        """ Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns
        :rtype: dict
        """
        return self._finalPatterns

    def printResults(self) -> None:
        """
        This function is used to print the results
        :return: None
        """
        print("Total number of Partial Periodic Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in ms:",  self.getRuntime())


if __name__ == "__main__":
    _ap = str()
    if len(_ab._sys.argv) == 5 or len(_ab._sys.argv) == 6:
        if len(_ab._sys.argv) == 6:
            _ap = PPP_ECLAT(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4], _ab._sys.argv[5])
        if len(_ab._sys.argv) == 5:
            _ap = PPP_ECLAT(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4])
        _ap.startMine()
        print("Total number of Partial Periodic Patterns:", len(_ap.getPatterns()))
        _ap.save(_ab._sys.argv[2])
        print("Total Memory in USS:", _ap.getMemoryUSS())
        print("Total Memory in RSS", _ap.getMemoryRSS())
        print("Total ExecutionTime in ms:", _ap.getRuntime())
    else:
        for i in [100, 200, 300, 400, 500]:
            _ap = PPP_ECLAT('/Users/tarunsreepada/Downloads/Temporal_T10I4D100K.csv', i, 5000, '\t')
            _ap.startMine()
            print("Total number of Maximal Partial Periodic Patterns:", len(_ap.getPatterns()))
            _ap.save('/Users/tarunsreepada/Downloads/output.txt')
            print(_ap.getPatternsAsDataFrame())
            print("Total Memory in USS:", _ap.getMemoryUSS())
            print("Total Memory in RSS", _ap.getMemoryRSS())
            print("Total ExecutionTime in ms:", _ap.getRuntime())
        print("Error! The number of input parameters do not match the total number of parameters provided")



import threading

class DfsThread(threading.Thread):
    """
       A thread class for performing DFS-based subgraph mining on a set of candidates.

       Args:
           graphDb (GraphDatabase): The graph database containing the input graphs.
           candidates (Queue): A queue containing candidate subgraphs to be mined.
           minSup (int): The minimum support threshold for frequent subgraph mining.
           tkgInstance (TKGInstance): An instance of the Top-K Graphs (TKG) algorithm for dynamic subgraph mining.
       """

    def __init__(self, graphDb, candidates, minSup, tkgInstance):
        threading.Thread.__init__(self)
        self.graphDb = graphDb
        self.candidates = candidates
        self.minSup = minSup
        self.tkgInstance = tkgInstance

    def run(self):
        """
            Runs the DFS-based subgraph mining process on the candidate subgraphs.

            This method is invoked when the thread is started using the `start()` method.

            Within the mining loop:
            - Retrieves a candidate subgraph from the candidates queue.
            - Checks if the candidate's support meets the minimum support threshold.
            - If the support is sufficient, invokes the gspanDynamicDFS method of the TKGInstance
              to perform dynamic DFS-based subgraph mining with the candidate's DFS code,
              the graph database, and the set of graph IDs associated with the candidate.

            This method continues to run until the candidates queue is empty or until a candidate
            with insufficient support is encountered.
            """
        while not self.candidates.empty():
            _, candidate = self.candidates.get()
            if len(candidate.setOfGraphsIds) < self.minSup:
                break
            self.tkgInstance.gspanDynamicDFS(candidate.dfsCode, self.graphDb, candidate.setOfGraphsIds)



# PPF_DFS is algorithm to mine the partial periodic frequent patterns.
#
# **Importing this algorithm into a python program**
#
#           from PAMI.partialPeriodicFrequentPattern.basic import PPF_DFS as alg
#
#           iFile = 'sampleTDB.txt'
#
#           minSup = 0.25 # can be specified between 0 and 1
#
#           maxPer = 300 # can  be specified between 0 and 1
#
#           minPR = 0.7 # can  be specified between 0 and 1
#
#           obj = alg.PPF_DFS(iFile, minSup, maxPer, minPR, sep)
#
#           obj.mine()
#
#           frequentPatterns = obj.getPatterns()
#
#           print("Total number of Frequent Patterns:", len(frequentPatterns))
#
#           obj.save(oFile)
#
#           Df = obj.getPatternInDataFrame()
#
#           memUSS = obj.getMemoryUSS()
#
#           print("Total Memory in USS:", memUSS)
#
#           memRSS = obj.getMemoryRSS()
#
#           print("Total Memory in RSS", memRSS)
#
#           run = obj.getRuntime()
#
#           print("Total ExecutionTime in seconds:", run)
#


__copyright__ = """
 Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
     Copyright (C)  2021 Rage Uday Kiran

"""


from PAMI.partialPeriodicFrequentPattern.basic.abstract import *
import deprecated
import numpy as np
import pandas as pd


class PPF_DFS(partialPeriodicPatterns):
    """
    **About this algorithm**

    :**Description**:   PPF_DFS is algorithm to mine the partial periodic frequent patterns.

    :**References**:    (Has to be added)

    :**parameters**:    - **iFile** (*str*) -- *Name of the Input file to mine complete set of correlated patterns.*
                        - **oFile** (*str*) -- *Name of the output file to store complete set of correlated patterns.*
                        - **minSup** (*int or float or str*) -- *The user can specify minSup either in count or proportion of database size. If the program detects the data type of minSup is integer, then it treats minSup is expressed in count.*
                        - **minPR** (*str*) -- *Controls the maximum number of transactions in which any two items within a pattern can reappear.*
                        - **maxPer** (*str*) -- *Controls the maximum number of transactions in which any two items within a pattern can reappear.*
                        - **sep** (*str*) -- *This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.*

    :**Attributes**:    - **memoryUSS** (*float*) -- *To store the total amount of USS memory consumed by the program.*
                        - **memoryRSS** (*float*) -- *To store the total amount of RSS memory consumed by the program.*
                        - **startTime** (*float*) -- *To record the start time of the mining process.*
                        - **endTime** (*float*) -- *To record the completion time of the mining process.*
                        - **minSup** (*int*) -- *The user given minSup.*
                        - **maxPer** (*int*) -- *The user given maxPer.*
                        - **minPR** (*int*) -- *The user given minPR.*
                        - **finalPatterns** (*dict*) -- *It represents to store the pattern.*

    :**Methods**:       - **mine()** -- *Mining process will start from here.*
                        - **Generation(prefix, itemsets, tidsets)** -- *Used to implement prefix class equibalence method to generate the periodic patterns recursively.*
                        - **getPartialPeriodicPatterns()** -- *Complete set of patterns will be retrieved with this function.*
                        - **storePatternsInFile(ouputFile)** -- *Complete set of frequent patterns will be loaded in to an output file.*
                        - **getPatternsAsDataFrame()** -- *Complete set of frequent patterns will be loaded in to an output file.*
                        - **getMemoryUSS()** -- *Total amount of USS memory consumed by the mining process will be retrieved from this function.*
                        - **getMemoryRSS()** -- *Total amount of RSS memory consumed by the mining process will be retrieved from this function.*
                        - **getRuntime()** -- *Total amount of runtime taken by the mining process will be retrieved from this function.*


    **Execution methods**

    **Terminal command**

    .. code-block:: console

      Format:

      (.venv) $ python3 PPF_DFS.py <inputFile> <outputFile> <minSup> <maxPer> <minPR>

      Example Usage:

      (.venv) $ python3 PPF_DFS.py sampleTDB.txt output.txt 0.25 300 0.7

    .. note:: minSup can be specified in support count or a value between 0 and 1.

    **Calling from a python program**

    .. code-block:: python

            from PAMI.partialPeriodicFrequentPattern.basic import PPF_DFS as alg

            iFile = 'sampleTDB.txt'

            minSup = 0.25 # can be specified between 0 and 1

            maxPer = 300 # can  be specified between 0 and 1

            minPR = 0.7 # can  be specified between 0 and 1

            obj = alg.PPF_DFS(inputFile, minSup, maxPer, minPR, sep)

            obj.mine()

            partialPeriodicFrequentPatterns = obj.getPatterns()

            print("Total number of partial periodic Patterns:", len(partialPeriodicFrequentPatterns))

            obj.save(oFile)

            Df = obj.getPatternInDf()

            memUSS = obj.getMemoryUSS()

            print("Total Memory in USS:", memUSS)

            memRSS = obj.getMemoryRSS()

            print("Total Memory in RSS", memRSS)

            run = obj.getRuntime()

            print("Total ExecutionTime in seconds:", run)

    **Credits**

    The complete program was written by Nakamura and revised by Tarun Sreepada under the supervision of Professor Rage Uday Kiran.

    """

    __path = ' '
    _partialPeriodicPatterns__iFile = ' '
    _partialPeriodicPatterns__oFile = ' '
    _partialPeriodicPatterns__sep = str()
    _partialPeriodicPatterns__minSup = str()
    _partialPeriodicPatterns__maxPer = str()
    _partialPeriodicPatterns__minPR = str()
    __tidlist = {}
    __last = 0
    __lno = 0
    __mapSupport = {}
    _partialPeriodicPatterns__finalPatterns = {}
    __runTime = float()
    _partialPeriodicPatterns__memoryUSS = float()
    _partialPeriodicPatterns__memoryRSS = float()
    _partialPeriodicPatterns__startTime = float()
    _partialPeriodicPatterns__endTime = float()
    __Database = []


    def _creatingItemSets(self) -> None:
        """
        Storing the complete transactions of the database/input file in a database variable

        :return: None
        """
        self._Database = []
        if isinstance(self._partialPeriodicPatterns__iFile, pd.DataFrame):
            data, ts = [], []
            if self._partialPeriodicPatterns__iFile.empty:
                print("its empty..")
            i = self._partialPeriodicPatterns__iFile.columns.values.tolist()
            if 'TS' in i:
                ts = self._partialPeriodicPatterns__iFile['TS'].tolist()
            if 'Transactions' in i:
                data = self._partialPeriodicPatterns__iFile['Transactions'].tolist()
            for i in range(len(data)):
                if data[i]:
                    tr = [str(ts[i])] + [x for x in data[i].split(self._partialPeriodicPatterns__sep)]
                    self._Database.append(tr)
                else:
                    self._Database.append([str(ts[i])])

        if isinstance(self._partialPeriodicPatterns__iFile, str):
            if validators.url(self._partialPeriodicPatterns__iFile):
                data = urlopen(self._partialPeriodicPatterns__iFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._partialPeriodicPatterns__sep)]
                    temp = [x for x in temp if x]
                    self._Database.append(temp)
            else:
                try:
                    with open(self._partialPeriodicPatterns__iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line.strip()
                            temp = [i.rstrip() for i in line.split(self._partialPeriodicPatterns__sep)]
                            temp = [x for x in temp if x]
                            self._Database.append(temp)
                except IOError:
                    print("File Not Found")
                    quit()
                    

    def _partialPeriodicPatterns__getPerSup(self, tids):
        """
        calculate ip of a pattern

        :param tids: tid list of the pattern
        :type tids: list
        :return: ip
        """
        # print(lno)
        tids = list(set(tids))
        tids.sort()
        per = 0
        sup = 0
        cur = 0
        if len(tids) == 0:
            return 0
        if abs(0 - tids[0]) <= self._partialPeriodicPatterns__maxPer:
            sup += 1
        for j in range(len(tids) - 1):
            i = j + 1
            per = abs(tids[i] - tids[j])
            if (per <= self._partialPeriodicPatterns__maxPer):
                sup += 1
        if abs(tids[len(tids) - 1] - self.__last) <= self._partialPeriodicPatterns__maxPer:
            sup += 1
        if sup == 0:
            return 0
        return sup


    def __convert(self, value):
        """
        to convert the type of user specified minSup value

        :param value: user specified minSup value
        :return: converted type
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (self._dbSize * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (self._dbSize * value)
            else:
                value = int(value)
        return value


    def startMine(self):
        self.mine()

    def _getPerSup(self, arr):
        """
        This function takes the arr as input and returns locs as output

        :param arr: an array contains the items.
        :type arr: array
        :return: locs
        """
        arr = list(arr)
        arr.append(self._maxTS)
        arr.append(0)
        arr = list(set(arr))
        arr = np.sort(arr)
        arr = np.diff(arr)

        locs = len(np.where(arr <= self._partialPeriodicPatterns__maxPer)[0])

        return locs
    
    def __recursive(self, cands, items):
        """
        This method processes candidate patterns, generates new candidates by intersecting
        itemsets, and filters them based on minimum support and periodic support ratio.
        If new candidates are found, the method recursively calls itself.

        :param cands: List of current candidate patterns.
        :type cands: List of tuple
        :param items: Dictionary where keys are candidate patterns and values are sets of transaction indices in which the pattern occurs.
        :type items: dict
        :return: None
        """

        for i in range(len(cands)):
            newCands = []
            nitems = {}
            for j in range(i + 1, len(cands)):
                intersection = items[cands[i]].intersection(items[cands[j]])
                if len(intersection) >= self._partialPeriodicPatterns__minSup:
                    perSup = self._getPerSup(intersection)
                    ratio = perSup / (len(intersection) + 1)
                    nCand = cands[i] + tuple([cands[j][-1]])
                    newCands.append(nCand)
                    nitems[nCand] = intersection
                    if ratio >= self._partialPeriodicPatterns__minPR:
                        self._partialPeriodicPatterns__finalPatterns[nCand] = [len(intersection), ratio]
            if len(newCands):
                self.__recursive(newCands, nitems)


    def mine(self):
        """
        Main program start with extracting the periodic frequent items from the database and
        performs prefix equivalence to form the combinations and generates closed periodic frequent patterns.
        """
        self._partialPeriodicPatterns__startTime = time.time()
        self._creatingItemSets()
        self._partialPeriodicPatterns__finalPatterns = {}
        
        items = {}
        tids = set()
        maxTS = 0
        for line in self._Database:
            index = int(line[0])
            tids.add(index)
            maxTS = max(maxTS, index)
            for item in line[1:]:
                if tuple([item]) not in items:
                    items[tuple([item])] = set()
                items[tuple([item])].add(index)

        self._maxTS = maxTS

        self._dbSize = maxTS

        self._partialPeriodicPatterns__minSup = self.__convert(self._partialPeriodicPatterns__minSup)
        self._partialPeriodicPatterns__maxPer = self.__convert(self._partialPeriodicPatterns__maxPer)
        self._partialPeriodicPatterns__minPR = float(self._partialPeriodicPatterns__minPR)

        cands = []
        nitems = {}

        for k, v in items.items():
            if len(v) >= self._partialPeriodicPatterns__minSup:
                perSup = self._getPerSup(v)
                cands.append(k)
                nitems[k] = v
                ratio = perSup / (len(v) + 1)
                if ratio >= self._partialPeriodicPatterns__minPR:
                    self._partialPeriodicPatterns__finalPatterns[k] = [len(v), ratio]

        self.__recursive(cands, nitems)

        temp = {}
        for k,v in self._partialPeriodicPatterns__finalPatterns.items():
            k = list(k)
            k = "\t".join(k)
            temp[k] = v
        self._partialPeriodicPatterns__finalPatterns = temp

        self._partialPeriodicPatterns__endTime = time.time()
        self.__runTime = self._partialPeriodicPatterns__endTime - self._partialPeriodicPatterns__startTime
        process = psutil.Process(os.getpid())
        self._partialPeriodicPatterns__memoryUSS = float()
        self._partialPeriodicPatterns__memoryRSS = float()
        self._partialPeriodicPatterns__memoryUSS = process.memory_full_info().uss
        self._partialPeriodicPatterns__memoryRSS = process.memory_info().rss

    def getMemoryUSS(self):
        """
        Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._partialPeriodicPatterns__memoryUSS

    def getMemoryRSS(self):
        """
        Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self._partialPeriodicPatterns__memoryRSS

    def getRuntime(self):
        """
        Calculating the total amount of runtime taken by the mining process

        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self.__runTime

    def getPatternsAsDataFrame(self):
        """
        Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        # print("Storing the patterns in a dataframe")
        dataFrame = {}
        data = []
        for a, b in self._partialPeriodicPatterns__finalPatterns.items():
            data.append([a, b[0], b[1]])
            dataFrame = pd.DataFrame(data, columns=['Patterns', 'Support', 'Periodic Ratio'])
        return dataFrame

    def save(self, outFile):
        """
        Complete set of frequent patterns will be loaded in to an output file

        :param outFile: name of the output file
        :type outFile: csv file
        """
        self.oFile = outFile
        with open(self.oFile, 'w') as f:
            for x, y in self._partialPeriodicPatterns__finalPatterns.items():
                # print(list(x), y)
                f.write(x + ":" + str(y[0]) + ":" + str(y[1]) + "\n")

    def getPatterns(self):
        """
        Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns
        :rtype: dict
        """
        return self._partialPeriodicPatterns__finalPatterns


    def printResults(self):
        """
        this function is used to print the results
        """
        print("Total number of Partial Periodic Frequent Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in s:", self.getRuntime())

if __name__ == '__main__':
    ap = str()
    if len(sys.argv) == 6 or len(sys.argv) == 7:
        if len(sys.argv) == 7:
            ap = PPF_DFS(sys.argv[1], sys.argv[3], sys.argv[4], sys.argv[5], sys.argv[6])
        if len(sys.argv) == 6:
            ap = PPF_DFS(sys.argv[1], sys.argv[3], sys.argv[4], sys.argv[5])
        ap.mine()
        print("Total number of Frequent Patterns:", len(ap.getPatterns()))
        ap.save(sys.argv[2])
        print("Total Memory in USS:", ap.getMemoryUSS())
        print("Total Memory in RSS", ap.getMemoryRSS())
        print("Total ExecutionTime in ms:", ap.getRuntime())
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")




# AprioriBitset is one of the fundamental algorithm to discover frequent patterns in a transactional database.
#
# **Importing this algorithm into a python program**
#
#             import PAMI.frequentPattern.basic.AprioriBitset as alg
#
#             iFile = 'sampleDB.txt'
#
#             minSup = 10  # can also be specified between 0 and 1
#
#             obj = alg.AprioriBitset(iFile, minSup)
#
#             obj.mine()
#
#             frequentPatterns = obj.getPatterns()
#
#             print("Total number of Frequent Patterns:", len(frequentPatterns))
#
#             obj.save(oFile)
#
#             Df = obj.getPatternInDataFrame()
#
#             memUSS = obj.getMemoryUSS()
#
#             print("Total Memory in USS:", memUSS)
#
#             memRSS = obj.getMemoryRSS()
#
#             print("Total Memory in RSS", memRSS)
#
#             run = obj.getRuntime()
#
#             print("Total ExecutionTime in seconds:", run)
#


__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from PAMI.frequentPattern.basic import abstract as _ab
from deprecated import deprecated


class Aprioribitset(_ab._frequentPatterns):
    """
    **About this algorithm**

    :**Description**:  AprioriBitset is one of the fundamental algorithm to discover frequent patterns in a transactional database.

    :**Reference**:  Mohammed Javeed Zaki: Scalable Algorithms for Association Mining. IEEE Trans. Knowl. Data Eng. 12(3):
                     372-390 (2000), https://ieeexplore.ieee.org/document/846291

    :**Parameters**:    - **iFile** (*str or URL or dataFrame*) -- *Name of the Input file to mine complete set of frequent patterns.*
                        - **oFile** (*str*) -- *Name of the output file to store complete set of frequent patterns.*
                        - **minSup** (*int or float or str*) -- *The user can specify minSup either in count or proportion of database size. If the program detects the data type of minSup is integer, then it treats minSup is expressed in count. Otherwise, it will be treated as float.*
                        - **sep** (*str*) -- *This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.*

    :**Attributes**:    - **startTime** (*float*) -- *To record the start time of the mining process.*
                        - **endTime** (*float*) -- *To record the completion time of the mining process.*
                        - **finalPatterns** (*dict*) -- *Storing the complete set of patterns in a dictionary variable.*
                        - **memoryUSS** (*float*) -- *To store the total amount of USS memory consumed by the program.*
                        - **memoryRSS** (*float*) -- *To store the total amount of RSS memory consumed by the program.*
                        - **Database** (*list*) -- *To store the transactions of a database in list.*



    **Execution methods**

    **Terminal command**

    .. code-block:: console

      Format:

      (.venv) $ python3 AprioriBitset.py <inputFile> <outputFile> <minSup>

      Example Usage:

      (.venv) $ python3 AprioriBitset.py sampleDB.txt patterns.txt 10.0

    .. note:: minSup can be specified  in support count or a value between 0 and 1.


    **Calling from a python program**

    .. code-block:: python

            import PAMI.frequentPattern.basic.Aprioribitset as alg

            iFile = 'sampleDB.txt'

            minSup = 10  # can also be specified between 0 and 1

            obj = alg.Aprioribitset(iFile, minSup)

            obj.mine()

            frequentPattern = obj.getPatterns()

            print("Total number of Frequent Patterns:", len(frequentPattern))

            obj.save(oFile)

            Df = obj.getPatternInDataFrame()

            memUSS = obj.getMemoryUSS()

            print("Total Memory in USS:", memUSS)

            memRSS = obj.getMemoryRSS()

            print("Total Memory in RSS", memRSS)

            run = obj.getRuntime()

            print("Total ExecutionTime in seconds:", run)


    **Credits**

    The complete program was written by Yudai Masu and revised by Tarun Sreepada under the supervision of Professor Rage Uday Kiran.

    """

    _startTime = float()
    _endTime = float()
    _finalPatterns = {}
    _iFile = " "
    _oFile = " "
    _sep = " "
    _minSup = str()
    _memoryUSS = float()
    _memoryRSS = float()
    _Database = []
    _mapSupport = {}
    _lno = 0

    def _convert(self, value):
        """
        To convert the user specified minSup value

        :param value: user specified minSup value
        :type value: int
        :return: converted type
        :rtype: int or float or string
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (len(self._Database) * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (len(self._Database) * value)
            else:
                value = int(value)
        return value

    def _creatingItemSets(self):
        """
        Storing the complete transactions of the database/input file in a database variable
        """
        self._Database = []
        self._mapSupport = {}
        if isinstance(self._iFile, _ab._pd.DataFrame):
            temp = []
            if self._iFile.empty:
                print("its empty..")
            i = self._iFile.columns.values.tolist()
            if 'Transactions' in i:
                self._Database = self._iFile['Transactions'].tolist()
                self._Database = [x.split(self._sep) for x in self._Database]
            else:
                print("The column name should be Transactions and each line should be separated by tab space or a seperator specified by the user")

        if isinstance(self._iFile, str):
            if _ab._validators.url(self._iFile):
                data = _ab._urlopen(self._iFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._sep)]
                    temp = [x for x in temp if x]
                    self._Database.append(temp)
            else:
                try:
                    with open(self._iFile, 'r') as f:
                        for line in f:
                            self._lno += 1
                            splitter = [i.rstrip() for i in line.split(self._sep)]
                            splitter = [x for x in splitter if x]
                            self._Database.append(splitter)
                except IOError:
                    print("File Not Found")
        self._minSup = self._convert(self._minSup)

    @deprecated("It is recommended to use 'mine()' instead of 'startMine()' for mining process. Starting from January 2025, 'startMine()' will be completely terminated.")

    def startMine(self):
        self.mine()

    def _bitPacker(self, data, maxIndex):
        """
        It takes the data and maxIndex as input and generates integer as output value.

        :param data: it takes data as input.
        :type data: int or float
        :param maxIndex: It converts the data into bits By taking the maxIndex value as condition.
        :type maxIndex: int
        """
        packed_bits = 0
        for i in data:
            packed_bits |= 1 << (maxIndex - i)

        return packed_bits

    def mine(self, memorySaver = True) -> None:
        """
        Frequent pattern mining process will start from here
        """
        self._startTime = _ab._time.time()

        self._Database = []

        self._creatingItemSets()

        items = {}
        index = 0
        for line in self._Database:
            for item in line:
                if tuple([item]) in items:
                    items[tuple([item])].append(index)
                else:
                    items[tuple([item])] = [index]
            index += 1

        # sort by length in descending order
        items = dict(sorted(items.items(), key=lambda x: len(x[1]), reverse=True))
        cands = []
        for key in items:
            if len(items[key]) >= self._minSup:
                self._finalPatterns[key] = len(items[key])
                cands.append(key)
                items[key] = self._bitPacker(items[key], index)
                # print(key, items[key])
            else:
                break

        if memorySaver:
            while cands:
                newCands = []
                for i in range(len(cands)):
                    for j in range(i + 1, len(cands)):
                        if cands[i][:-1] == cands[j][:-1]:
                            newCand = tuple(cands[i] + tuple([cands[j][-1]]))
                            intersection = items[tuple([newCand[0]])]
                            for k in range(1, len(newCand)):
                                intersection &= items[tuple([newCand[k]])]
                            count = int.bit_count(intersection)
                            if count >= self._minSup:
                                newCands.append(newCand)
                                self._finalPatterns[newCand] = count
                        else:
                            break

                cands = newCands
        else:
            while cands:
                newCands = []
                for i in range(len(cands)):
                    for j in range(i + 1, len(cands)):
                        if cands[i][:-1] == cands[j][:-1]:
                            newCand = tuple(cands[i] + tuple([cands[j][-1]]))
                            # intersection = items[tuple([newCand[0]])]
                            # for k in range(1, len(newCand)):
                            #     intersection &= items[tuple([newCand[k]])]
                            intersection = items[cands[i]] & items[cands[j]]
                            count = int.bit_count(intersection)
                            if count >= self._minSup:
                                newCands.append(newCand)
                                self._finalPatterns[newCand] = count
                                items[newCand] = intersection
                        else:
                            break

                cands = newCands

        self._endTime = _ab._time.time()
        process = _ab._psutil.Process(_ab._os.getpid())
        self._memoryUSS = float()
        self._memoryRSS = float()
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss
        print("Frequent patterns were generated successfully using Apriori algorithm ")

    def getMemoryUSS(self):
        """
        Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self):
        """
        Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryRSS

    def getRuntime(self):
        """
        Calculating the total amount of runtime taken by the mining process

        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self._endTime - self._startTime

    def getPatternsAsDataFrame(self) -> _ab._pd.DataFrame:
        """
        Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame

        """

        # time = _ab._time.time()
        # dataFrame = {}
        # data = []
        # for a, b in self._finalPatterns.items():
        #     # data.append([a.replace('\t', ' '), b])
        #     data.append([" ".join(a), b])
        #     dataFrame = _ab._pd.DataFrame(data, columns=['Patterns', 'Support'])
        # print("Time taken to convert the frequent patterns into DataFrame is: ", _ab._time.time() - time)

        dataFrame = _ab._pd.DataFrame(list([[" ".join(x), y] for x,y in self._finalPatterns.items()]), columns=['Patterns', 'Support'])
        # dataFrame = _ab._pd.DataFrame(list(self._finalPatterns.items()), columns=['Patterns', 'Support'])

        return dataFrame

    def save(self, outFile: str, seperator = "\t" ) -> None:
        """
        Complete set of frequent patterns will be loaded in to an output file

        :param outFile: name of the output file
        :type outFile: csvfile
        :param seperator: variable to store separator value
        :type seperator: string
        :return: None
        """

        # self._oFile = outFile
        # writer = open(self._oFile, 'w+')
        # for x, y in self._finalPatterns.items():
        #     patternsAndSupport = x.strip() + ":" + str(y[0])
        #     writer.write("%s \n" % patternsAndSupport)
        with open(outFile, 'w') as f:
            for x, y in self._finalPatterns.items():
                x = seperator.join(x)
                f.write(f"{x}:{y}\n")

    def getPatterns(self):
        """
        Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns
        :rtype: dict
        """
        return self._finalPatterns

    def printResults(self):
        """
        This function is used to print the result
        """
        print("Total number of Frequent Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in ms:", self.getRuntime())


if __name__ == "__main__":
    _ap = str()
    if len(_ab._sys.argv) == 4 or len(_ab._sys.argv) == 5:
        if len(_ab._sys.argv) == 5:
            _ap = Aprioribitset(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4])
        if len(_ab._sys.argv) == 4:
            _ap = Aprioribitset(_ab._sys.argv[1], _ab._sys.argv[3])
        _ap.startMine()
        _ap.mine()
        print("Total number of Frequent Patterns:", len(_ap.getPatterns()))
        _ap.save(_ab._sys.argv[2])
        print("Total Memory in USS:", _ap.getMemoryUSS())
        print("Total Memory in RSS", _ap.getMemoryRSS())
        print("Total ExecutionTime in ms:", _ap.getRuntime())
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")


    obj = Aprioribitset("/Users/tarunsreepada/Downloads/Transactional_T10I4D100K.csv", 75)
    obj.mine(memorySaver=True)
    print(obj.printResults())


# generateTemporalDatabase is a code used to convert the database into Temporal database.
#
#  **Importing this algorithm into a python program**
#  --------------------------------------------------------
#
#             from PAMI.extras.generateDatabase import generateTemporalDatabase as db
#
#             obj = db.generateTemporalDatabase(100, 10, 6, oFile, %, "\t")
#
#             obj.save()
#
#             obj.getFileName("outputFileName") # to create a file
#
#             obj.getDatabaseAsDataFrame("outputFileName") # to convert database into dataframe
#
#             obj.createTemporalFile("outputFileName") # to get outputfile
#




__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Tuple, List, Union
import pandas as pd
import numpy as np
import random
import sys
import os

class generateTemporalDatabase:
    """
    :Description:   generateTemporalDatabase creates a temporal database and outputs a database or a frame depending on input

    :Attributes:
        :param numOfTransactions: int
            number of transactions
        :param avgLenOfTransactions: int
            average length of transactions
        :param numItems: int
            number of items
        :param outputFile: str
            output file name
        :param percentage: int
            percentage of coinToss for TID of temporalDatabase
        :param sep: str
            seperator for database output file
        :param typeOfFile: str
            specify database or dataframe to get corresponding output

    :Methods:
        getFileName():
            returns filename
        createTemporalFile():
            creates temporal database file or dataframe
        getDatabaseAsDataFrame:
            returns dataframe
        performCoinFlip():
            Perform a coin flip with the given probability
        tuning():
            Tune the arrayLength to match avgLenOfTransactions
        createTemporalFile():
            create Temporal database or dataframe depending on input

    **Importing this algorithm into a python program**
    --------------------------------------------------------
    .. code-block:: python

            from PAMI.extras.generateDatabase import generateTemporalDatabase as db

            numOfTransactions = 100
            numItems = 15
            avgTransactionLength = 6
            outFileName = 'temporal_ot.txt'
            sep = '\t'
            percent = 75
            frameOrBase = "dataframe" # if you want to get dataframe as output
            frameOrBase = "database" # if you want to get database/csv/file as output

            temporalDB = db.generateTemporalDatabase(numOfTransactions, avgTransactionLength, numItems, outFileName, percent, sep, frameOrBase )
            temporalDB.createTemporalFile()
            print(temporalDB.getDatabaseAsDataFrame())

    """
    def __init__(self, numOfTransactions: int, avgLenOfTransactions: int, 
                 numItems: int, outputFile: str, percentage: int=50,
                 sep: str='\t', typeOfFile: str="Database") -> None:
        
        """
        :Description:   Initialize the generateTemporalDatabase class

        :Attributes:
            :param numOfTransactions: int
                number of transactions
            :param avgLenOfTransactions: int
                average length of transactions
            :param numItems: int
                number of items
            :param outputFile: str
                output file name
            :param percentage: int
                percentage of coinToss for TID of temporalDatabase
            :param sep: str
                seperator for database output file
            :param typeOfFile: str
                specify database or dataframe to get corresponding output

        :Methods:
            getFileName():
                returns filename
            createTemporalFile():
                creates temporal database file or dataframe
            getDatabaseAsDataFrame:
                returns dataframe
            performCoinFlip():
                Perform a coin flip with the given probability
            tuning():
                Tune the arrayLength to match avgLenOfTransactions
            createTemporalFile():
                create Temporal database or dataframe depending on input
        
        """

        self.numOfTransactions = numOfTransactions
        self.avgLenOfTransactions = avgLenOfTransactions
        self.numItems = numItems
        self.outputFile = outputFile
        if percentage > 1:
            self.percentage = percentage / 100
        else:
            self.percentage = percentage
        self.sep = sep
        self.typeOfFile = typeOfFile.lower()

    def getFileName(self) -> str:
        """
        return filename
        :return: filename
        :rtype: str
        """
        return self.outputFile

    def getDatabaseAsDataFrame(self) -> pd.DataFrame:
        """
        return dataframe
        :return: dataframe
        :rtype: pd.DataFrame
        """
        return self.df
    
    def performCoinFlip(self, probability: float) -> bool:
        """
        Perform a coin flip with the given probability.
        :param probability: probability to perform coin flip
        :type probability: float
        :return: True if coin flip is performed, False otherwise
        :rtype: bool
        """
        result = np.random.choice([0, 1], p=[1 - probability, probability])
        return result == 1


    def tuning(self, array, sumRes) -> list:
        """
        Tune the array so that the sum of the values is equal to sumRes

        :param array: list of values

        :type array: list

        :param sumRes: target sum

        :type sumRes: int

        :return: list of values with the sum equal to sumRes after tuning

        :rtype: list
        """

        # first generate a random array of length n whose values average to m
        values = np.random.randint(1, self.numItems, len(array))

        while np.sum(values) != sumRes:
            # get index of largest value
            # if sum is too large, decrease the largest value
            if np.sum(values) > sumRes:
                maxIndex = np.argmax(values)
                values[maxIndex] -= 1
            # if sum is too small, increase the smallest value
            else:
                minIndex = np.argmin(values)
                values[minIndex] += 1

        # get location of all values greater than numItems
        
        for i in range(len(array)):
            array[i][1] = values[i]

        return array

    def createTemporalFile(self) -> None:
        """
        create Temporal database or dataframe depending on input
        :return: None
        """

        db = []
        lineSize = []
        for i in range(self.numOfTransactions):
            db.append([i])
            if self.performCoinFlip(self.percentage):
                lineSize.append([i,0])
        
        # make it so that sum of lineSize[1] equal to numTransactions * avgLenOfTransactions
        sumRes = self.numOfTransactions * self.avgLenOfTransactions
        self.tuning(lineSize, sumRes)

        for i in range(len(lineSize)):
            if lineSize[i][1] > self.numItems:
                raise ValueError("Error: Either increase numItems or decrease avgLenOfTransactions or modify percentage")
            line = np.random.choice(range(1, self.numItems + 1), lineSize[i][1], replace=False)
            db[lineSize[i][0]].extend(line)

        if self.typeOfFile == "database":
            with open(self.outputFile, "w") as outFile:
                for line in db:
                    outFile.write(self.sep.join(map(str, line)) + '\n')
            outFile.close()

        if self.typeOfFile == "dataframe":
            data = {
                'timestamp': [line[0] for line in db],
                'transactions': pd.Series([line[1:] for line in db])
            }
            self.df = pd.DataFrame(data)

        print("Temporal database created successfully")


if __name__ == '__main__':
    numOfTransactions = 100
    numItems = 20
    avgTransactionLength = 6
    outFileName = 'temporal_out.txt'
    sep = '\t'
    frameOrBase = "database"

    temporalDB = generateTemporalDatabase(numOfTransactions, avgTransactionLength, numItems, outFileName)

    temporalDB.createTemporalFile()

    numOfTransactions = 100
    numItems = 15
    avgTransactionLength = 6
    outFileName = 'temporal_ot.txt'
    sep = '\t'
    percent = 75
    frameOrBase = "dataframe"

    temporalDB = generateTemporalDatabase(numOfTransactions, avgTransactionLength, numItems, outFileName, percent, sep, frameOrBase )
    temporalDB.createTemporalFile()
    print(temporalDB.getDatabaseAsDataFrame())

    obj = generateTemporalDatabase(sys.argv[1], sys.argv[2], sys.argv[3], sys.argv[4])
    obj.createTemporalFile(sys.argv[5])



class Edge:
    def __init__(self, v1, v2, edgeLabel, existenceProbability=1.0):
        self.v1 = v1
        self.v2 = v2
        self.edgeLabel = edgeLabel
        self.existenceProbability = existenceProbability

    def getEdgeLabel(self):
        return self.edgeLabel

    def getExistenceProbability(self):
        return self.existenceProbability

    def __hash__(self):
        return hash((self.v1, self.v2, self.edgeLabel))

    def __eq__(self, other):
        if not isinstance(other, Edge):
            return False
        return (self.v1 == other.v1 and
                self.v2 == other.v2 and
                self.edgeLabel == other.edgeLabel)

    def __repr__(self):
        return f"Edge(v1: {self.v1}, v2: {self.v2}, Label: {self.edgeLabel})"




import random as _rd


class CreateSyntheticTemporal:
    """
    This class creates a synthetic temporal database.

    Attributes:
        total_transactions (int): Number of transactions.
        num_of_items (int): Number of items.
        avg_transaction_length (int): The length of average transaction.

    Methods:
        create_temporal_database(output_file)
            Create temporal database and store it in the specified output file.

    Credits:
        The complete program was written by A.Hemanth sree sai under the supervision of Professor Rage Uday Kiran.
    """

    def __init__(self, total_transactions: int, num_of_items: int, avg_transaction_length: int) -> None:
        """
        Constructor to initialize the database parameters.

        Parameters:
            total_transactions (int): Number of transactions.
            num_of_items (int): Number of items.
            avg_transaction_length (int): The length of average transaction.
        """
        self.total_transactions = total_transactions
        self.num_of_items = num_of_items
        self.avg_transaction_length = avg_transaction_length

    def create_temporal_database(self, output_file: str) -> None:
        """
        Create temporal database and store it in the specified output file.

        Parameters:
            output_file (str): File name or path to store the database.
        """
        if self.avg_transaction_length > self.num_of_items:
            print("Error: avg_transaction_length cannot exceed num_of_items.")
            return

        count = 1
        with open(output_file, 'w') as writer:
            for _ in range(self.total_transactions):
                length = _rd.randint(1, self.avg_transaction_length + 20)
                st = str(count) + '\t'
                random_numbers = self.generate_random_numbers(13, 2000)

                # Checking if avgTransactionLength exceeds numOfItems
                if self.avg_transaction_length > self.num_of_items:
                    print("Error: avg_transaction_length cannot exceed num_of_items.")
                    return

                for _ in range(length):
                    item = _rd.randint(1, self.num_of_items)
                    st = st + str(item) + '\t'
                writer.write("%s \n" % st)
                count += 1

    def generate_random_numbers(self, n: int, target_sum: int) -> list[float]:
        """
        Generate a list of random numbers with a specified target sum.

        Parameters:
            n (int): Number of random numbers to generate.
            target_sum (int): Target sum for the generated random numbers.

        Returns:
            list: List of generated random numbers normalized and multiplied by the target sum.
        """
        rand_numbers = [_rd.uniform(0, 1) for _ in range(n)]
        rand_sum = sum(rand_numbers)
        normalized_numbers = [num / rand_sum for num in rand_numbers]
        result = [round(num * target_sum) for num in normalized_numbers]
        return result


if __name__ == "__main__":
    ap = CreateSyntheticTemporal(100000, 870, 10)
    ap.create_temporal_database("temporal_T10.csv")
else:
    print("Error! The number of input parameters does not match the total number of parameters provided")



#  Usage
#  obj = ConvertFormat('iFile', 'oFile')
#
class ConvertFormat:
    def __init__(self, iFile, oFile):
        self.iFile = iFile
        self.oFile = oFile
        self.convert()

    def _writeGraphToFile(self, graph, oFile):
        node_str = ' '.join(f"{node} {label}" for node, label in sorted(graph['nodes']))
        edge_str = ' '.join(f"{u} {v} {label}" for u, v, label in graph['edges'])
        oFile.write(f"{node_str} : {edge_str}\n")

    def convert(self):
        graph = {}
        with open(self.iFile, 'r') as iFile, open(self.oFile, 'w') as oFile:
            for line in iFile:
                parts = line.strip().split()
                if not parts:
                    continue
                if parts[0] == 't':
                    if graph:
                        self._writeGraphToFile(graph, oFile)
                    graph = {'nodes': [], 'edges': []}
                elif parts[0] == 'v':
                    graph['nodes'].append((int(parts[1]), parts[2]))
                elif parts[0] == 'e':
                    graph['edges'].append((int(parts[1]), int(parts[2]), parts[3]))
            if graph:
                self._writeGraphToFile(graph, oFile)


# DenseFormatDF in this code the dense dataframe is converting databases into different transactional, temporal, utility types.
#
# **Importing this algorithm into a python program**
# --------------------------------------------------------
#
#             from PAMI.extras.DF2DB import DenseFormatDF as db
#
#             obj = db.DenseFormatDF(idf, ">=", 16)
#
#             obj.save(oFile)
#
#             obj.convert2TransactionalDatabase("outputFileName") # To create transactional database
#
#             obj.convert2TemporalDatabase("outputFileName") # To create temporal database
#
#             obj.convert2MultipleTimeSeries("outputFileName") # To create Mutliple TimeSeries database
#
#             obj.convert2UtilityDatabase("outputFileName") # To create utility database
#
#             obj.getFileName("outputFileName") # To get file name of the database
#





__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import operator
from typing import Union

condition_operator = {
    '<': operator.lt,
    '>': operator.gt,
    '<=': operator.le,
    '>=': operator.ge,
    '==': operator.eq,
    '!=': operator.ne
}


class DenseFormatDF:
    """
    :Description: This class create Data Base from DataFrame.

    :Attributes:

        :param inputDF: dataframe :
            It is dense DataFrame
        :param condition: str :
            It is condition to judge the value in dataframe
        :param thresholdValue: int or float :
            User defined value.


    **Importing this algorithm into a python program**
    --------------------------------------------------------
    .. code-block:: python

            from PAMI.extras.DF2DB import DenseFormatDF as db

            obj = db.DenseFormatDF(iDdf, ">=", 16 )

            obj.convert2TransactionalDatabase("outputFileName") # To create transactional database

            obj.convert2TemporalDatabase("outputFileName") # To create temporal database

            obj.convert2MultipleTimeSeries("outputFileName") # To create Multiple TimeSeries database

            obj.convert2UtilityDatabase("outputFileName") # To create utility database

            obj.getFileName("outputFileName") # To get file name of the database
    """

    def __init__(self, inputDF) -> None:
        self.inputDF = inputDF
        self.tids = []
        self.items = []
        self.outputFile = ' '
        self.items = list(self.inputDF.columns.values)
        self.tids = list(self.inputDF.index)

    def convert2TransactionalDatabase(self, outputFile: str, condition: str, thresholdValue: Union[int, float]) -> None:
        """
        :Description: Create transactional data base

        :Attributes:

             :param outputFile: Write transactional database into outputFile

             :type outputFile: str

             :param condition: It is condition to judge the value in dataframe

             :type condition: str

             :param thresholdValue: User defined value.

             :type thresholdValue: Union[int, float]
        """


        self.outputFile = outputFile
        with open(outputFile, 'w') as f:
            if condition not in condition_operator:
                print('Condition error')
            else:
                for tid in self.tids:
                    transaction = [item for item in self.items if
                                   condition_operator[condition](self.inputDF.at[tid, item], thresholdValue)]
                    if len(transaction) > 1:
                        f.write(f'{transaction[0]}')
                        for item in transaction[1:]:
                            f.write(f'\t{item}')
                    elif len(transaction) == 1:
                        f.write(f'{transaction[0]}')
                    else:
                        continue
                    f.write('\n')

    def convert2TemporalDatabase(self, outputFile: str, condition: str, thresholdValue: Union[int, float]) -> None:
        """
        :Description: Create temporal database

        :param outputFile: Write temporal database into outputFile

        :type outputFile: str

        :param condition: It is condition to judge the value in dataframe

        :type condition: str

        :param thresholdValue: User defined value.

        :type thresholdValue: Union
        """

        self.outputFile = outputFile
        with open(outputFile, 'w') as f:
            if condition not in condition_operator:
                print('Condition error')
            else:
                for tid in self.tids:
                    transaction = [item for item in self.items if
                                   condition_operator[condition](self.inputDF.at[tid, item], thresholdValue)]
                    if len(transaction) > 1:
                        f.write(f'{tid + 1}')
                        for item in transaction:
                            f.write(f'\t{item}')
                    elif len(transaction) == 1:
                        f.write(f'{tid + 1}')
                        f.write(f'\t{transaction[0]}')
                    else:
                        continue
                    f.write('\n')

    def convert2MultipleTimeSeries(self, interval: int, outputFile: str, condition: str,
                                   thresholdValue: Union[int, float]) -> None:
        """
        :Description: Create the multiple time series database.

        :param outputFile:  Write multiple time series database into outputFile.

        :type outputFile:  str

        :param interval: Breaks the given timeseries into intervals.

        :type interval: int

        :param condition: It is condition to judge the value in dataframe

        :param thresholdValue: User defined value.

        :type thresholdValue: int or float
        """
        self.outputFile = outputFile
        writer = open(self.outputFile, 'w+')
        # with open(self.outputFile, 'w+') as f:
        count = 0
        tids = []
        items = []
        values = []
        for tid in self.tids:
            count += 1
            transaction = [item for item in self.items if
                           condition_operator[condition](self.inputDF.at[tid, item], thresholdValue)]
            for i in transaction:
                tids.append(count)
                items.append(i)
                values.append(self.inputDF.at[tid, i])
            if count == interval:
                s1, s, ss = str(), str(), str()
                if len(values) > 0:

                    for j in range(len(tids)):
                        s1 = s1 + str(tids[j]) + '\t'
                    for j in range(len(items)):
                        s = s + items[j] + '\t'
                    for j in range(len(values)):
                        ss = ss + str(values[j]) + '\t'

                s2 = s1 + ':' + s + ':' + ss
                writer.write("%s\n" % s2)
                tids, items, values = [], [], []
                count = 0

    def convert2UncertainTransactional(self, outputFile: str, condition: str,
                                       thresholdValue: Union[int, float]) -> None:
        self.outputFile = outputFile
        with open(outputFile, 'w') as f:
            if condition not in condition_operator:
                print('Condition error')
            else:
                for tid in self.tids:
                    transaction = [item for item in self.items if
                                   condition_operator[condition](self.inputDF.at[tid, item], thresholdValue)]
                    uncertain = [self.inputDF.at[tid, item] for item in self.items if
                                 condition_operator[condition](self.inputDF.at[tid, item], thresholdValue)]
                    if len(transaction) > 1:
                        f.write(f'{transaction[0]}')
                        for item in transaction[1:]:
                            f.write(f'\t{item}')
                        f.write(f':')
                        for value in uncertain:
                            tt = 0.1 + 0.036 * abs(25 - value)
                            tt = round(tt, 2)
                            f.write(f'\t{tt}')
                    elif len(transaction) == 1:
                        f.write(f'{transaction[0]}')
                        tt = 0.1 + 0.036 * abs(25 - uncertain[0])
                        tt = round(tt, 2)
                        f.write(f':{tt}')
                    else:
                        continue
                    f.write('\n')

    def convert2UtilityDatabase(self, outputFile: str) -> None:
        """
        :Description: Create the utility database.

        :param outputFile: Write utility database into outputFile

        :type outputFile: str

        :return: None
        """

        self.outputFile = outputFile
        with open(self.outputFile, 'w') as f:
            for tid in self.tids:
                df = self.inputDF.loc[tid].dropna()
                f.write(f'{df.index[0]}')
                for item in df.index[1:]:
                    f.write(f'\t{item}')
                f.write(f':{df.sum()}:')
                f.write(f'{df.at[df.index[0]]}')

                for item in df.index[1:]:
                    f.write(f'\t{df.at[item]}')
                f.write('\n')

    def getFileName(self) -> str:
        """
        :return: outputFile name
        :rtype: str
        """

        return self.outputFile

# Dataframes do not run from a terminal

# if __name__ == '__main__':
#     obj = DenseFormatDF(sys.argv[1], sys.argv[2], sys.argv[3])
#     obj.convert2TransactionalDatabase(sys.argv[4])
#     transactionalDB = obj.getFileName()
#     print(transactionalDB)


# TransactionalDatabase is a collection of transactions. It only considers the data in  transactions and ignores the metadata.
#
#  **Importing this algorithm into a python program**
#  --------------------------------------------------------
#     from PAMI.extras.syntheticDataGenerator import TransactionalDatabase as db
#
#     obj = db(10, 5, 10)
#
#     obj.create()
#
#     obj.save('db.txt')
#
#     print(obj.getTransactions())
#

import numpy as np
import pandas as pd
import sys,psutil,os,time


__copyright__ = """
 Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

class TransactionalDatabase:
    """
    :Description: TransactionalDatabase is a collection of transactions. It only considers the data in  transactions and ignores the metadata.
    :Attributes:

        numLines: int
            Number of lines
        avgItemsPerLine: int
            Average number of items per line
        numItems: int
            Total number of items
        memoryUSS : float
            To store the total amount of USS memory consumed by the program
        memoryRSS : float
                    To store the total amount of RSS memory consumed by the program
        startTime : float
                    To record the start time of the mining process
        endTime : float
                    To record the completion time of the mining process

    :Methods:

        create: 
            Generate the transactional database
        save: 
            Save the transactional database to a user-specified file
        getTransactions: 
            Get the transactional database
        getMemoryUSS()
            Total amount of USS memory consumed by the mining process will be retrieved from this function
        getMemoryRSS()
            Total amount of RSS memory consumed by the mining process will be retrieved from this function
        getRuntime()
            Total amount of runtime taken by the mining process will be retrieved from this function

    **Methods to execute code on terminal**
    ---------------------------------------------

    .. code-block:: console

      Format:

      (.venv) $ python3 TransactionalDatabase.py <numLines> <avgItemsPerLine> <numItems>

      Example Usage:

      (.venv) $ python3 TransactionalDatabase.py 50.0 10.0 100



    **Importing this algorithm into a python program**
    --------------------------------------------------------
        from PAMI.extras.syntheticDataGenerator import TransactionalDatabase as db

        obj = db.TransactionalDatabase(10, 5, 10)

        obj.create()

        obj.save('db.txt')

        print(obj.getTransactions())

    
    """

    def __init__(self, databaseSize, avgItemsPerTransaction, numItems,sep = "\t") -> None:
        """
        Initialize the transactional database with the given parameters

        :param databaseSize: total number of transactions in the database
        :type databaseSize: int
        :param avgItemsPerTransaction: average number of items per transaction
        :type avgItemsPerTransaction: int
        :param numItems: total number of items
        :type numItems: int
        :param sep: separator to distinguish the items in a transaction
        :type sep: str
        """

        self.databaseSize = databaseSize
        self.avgItemsPerTransaction = avgItemsPerTransaction
        self.numItems = numItems
        self.sep = sep
        self.db = []
        self._startTime = float()
        self._endTime = float()
        self._memoryUSS = float()
        self._memoryRSS = float()
    def _generateArray(self, nums, avg, maxItems) -> list:
        """
        Generate a random array of length n whose values average to m

        :param nums: number of values
        :type nums: int
        :param avg: average value
        :type avg: int
        :param maxItems: maximum value
        :type maxItems: int

        Returns:
        values: list - random array
        """

        # generate n random values
        values = np.random.randint(1, avg * 2, nums)
        sums = np.sum(values)
        weights = values / sums

        # Calculate sumRes
        sumRes = nums * avg

        # Adjust values based on weights and sumRes
        new_values = np.round(sumRes * weights).astype(int)

        # if all transactions have 0 items, add 1 item to each transaction
        for loc in np.where(new_values < 1)[0]:
            new_values[loc] += 1

        difference = sumRes - np.sum(new_values)
        if difference > 0:
            for i in range(difference):
                index = np.random.randint(0, len(new_values))
                new_values[index] += 1
        else:
            for i in range(abs(difference)):
                index = np.random.randint(0, len(new_values))
                new_values[index] -= 1

        return values

    def create(self) -> None:
        """
        Generate the transactional database with the given input parameters.
        Returns: None
        """
        self._startTime = time.time()
        values = self._generateArray(self.databaseSize, self.avgItemsPerTransaction, self.numItems)

        self.db = []
        for i in range(self.databaseSize):
            self.db.append(np.random.choice(range(1, self.numItems + 1), values[i], replace=False))

    def save(self, filename) -> None:
        """
        Save the transactional database to a file

        :param filename: name of the file
        :type filename: str
        """

        with open(filename, 'w') as f:
            for line in self.db:
                f.write(str(self.sep).join(map(str, line)) + '\n')

    def getTransactions(self, sep = "\t") -> pd.DataFrame:
        """
        Get the transactional database in dataFrame format

        Returns:
        db: pd.dataFrame - transactional database
        """
        column = "Transactions"
        db = pd.DataFrame(columns=[column])
        db[column] = [sep.join(map(str, line)) for line in self.db]
        return db

    def getMemoryUSS(self) -> float:
        """
        Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """
        process = psutil.Process(os.getpid())
        self._memoryUSS = process.memory_full_info().uss
        return self._memoryUSS

    def getMemoryRSS(self) -> float:
        """
        Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """
        process = psutil.Process(os.getpid())
        self._memoryRSS = process.memory_info().rss
        return self._memoryRSS

    def getRuntime(self) -> float:
        """
        Calculating the total amount of runtime taken by the mining process


        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """
        self._endTime = time.time()
        return self._endTime - self._startTime
if __name__ == "__main__":

    if len(sys.argv) == 5:
        obj = TransactionalDatabase(int(sys.argv[1]), int(sys.argv[2]), int(sys.argv[3]))
        obj.create()
        obj.save(sys.argv[4])
        print("Total Memory in USS:", obj.getMemoryUSS())
        print("Total Memory in RSS", obj.getMemoryRSS())
        print("Total ExecutionTime in ms:", obj.getRuntime())
    if len(sys.argv) == 6:
        obj = TransactionalDatabase(int(sys.argv[1]), int(sys.argv[2]), int(sys.argv[3]), sys.argv[4])
        obj.create()
        obj.save(sys.argv[5])
        print("Total Memory in USS:", obj.getMemoryUSS())
        print("Total Memory in RSS", obj.getMemoryRSS())
        print("Total ExecutionTime in ms:", obj.getRuntime())
    else:
        raise ValueError("Invalid number of arguments. Args: <numLines> <avgItemsPerLine> <numItems> <filename> or Args: <numLines> <avgItemsPerLine> <numItems> <sep> <filename>")
    


# results2Latex is used to convert the given results into LatexFile.
#
# **Importing this algorithm into a python program**
#
#             from PAMI.extras.graph import results2Latex
#
#             obj = results2Latex()
#
#             obj.print(resultsDF, xaxis='minSup', yaxis='patterns', label='algorithm')
#
#             obj.print(resultsDF, xaxis='minSup', yaxis='runtime', label='algorithm')
#
#             obj.print(resultsDF, xaxis='minSup', yaxis='memoryRSS', label='algorithm')
#
#             obj.print(resultsDF, xaxis='minSup', yaxis='memoryUSS', label='algorithm')
#
#             obj.save(resultsDF, xaxis='minSup', yaxis='patterns', label='algorithm', oFile='patterns.tex')
#
#             obj.save(resultsDF, xaxis='minSup', yaxis='runtime', label='algorithm', oFile='runtime.tex')
#
#             obj.save(resultsDF, xaxis='minSup', yaxis='memoryRSS', label='algorithm', oFile='memoryRSS.tex')
#
#             obj.save(resultsDF, xaxis='minSup', yaxis='memoryUSS', label='algorithm', oFile='memoryUSS.tex')
#

__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""



import pandas as pd

class results2Latex:
    """
    **About this algorithm**

    :**Description**:  generateLatexCode is used to convert the given DataFrame into a LaTeX file.

    :**Reference**:

    **Attributes:** - **latexCode** (*str*) -- Stores the generated LaTeX code.

    **Methods:**    - **generateLatexCode(result: pd.DataFrame, xColumn, yColumn, algorithmColumn=None)** -- *Generates LaTeX code based on DataFrame columns.*

                    - **print(result: pd.DataFrame, xaxis, yaxis, label=None)** -- *Prints the LaTeX code.*

                    - **save(result: pd.DataFrame, xaxis, yaxis, label=None, oFile="output.tex")** -- *Saves the LaTeX code to a .tex file.*

    **Importing this algorithm into a Python program**

    .. code-block:: python

            from PAMI.extras.graph import results2Latex

            resultsDF = Dataframe

            obj = results2Latex()

             obj.print(resultsDF, xaxis='minSup', yaxis='patterns', label='algorithm')

             obj.print(resultsDF, xaxis='minSup', yaxis='runtime', label='algorithm')

             obj.print(resultsDF, xaxis='minSup', yaxis='memoryRSS', label='algorithm')

             obj.print(resultsDF, xaxis='minSup', yaxis='memoryUSS', label='algorithm')

             obj.save(resultsDF, xaxis='minSup', yaxis='patterns', label='algorithm', oFile='patterns.tex')

             obj.save(resultsDF, xaxis='minSup', yaxis='runtime', label='algorithm', oFile='runtime.tex')

             obj.save(resultsDF, xaxis='minSup', yaxis='memoryRSS', label='algorithm', oFile='memoryRSS.tex')

             obj.save(resultsDF, xaxis='minSup', yaxis='memoryUSS', label='algorithm', oFile='memoryUSS.tex')

"""

    def __init__(self):
        self.latexCode = ""

    def generateLatexCode(self, result: pd.DataFrame, xColumn, yColumn, algorithmColumn=None) -> None:
        if result.empty:
            raise ValueError("The input DataFrame is empty. Please provide a DataFrame with data.")

        requiredColumns = [xColumn, yColumn]
        if algorithmColumn:
            requiredColumns.append(algorithmColumn)
        for col in requiredColumns:
            if col not in result.columns:
                raise ValueError(f"Column '{col}' does not exist in the DataFrame.")

        titles = [xColumn, yColumn]
        legendary = pd.unique(result[algorithmColumn] if algorithmColumn else result.iloc[:, 0].values.ravel())
        xaxisValues = result[xColumn].values
        yaxisValues = result[yColumn].values
        algo = result[algorithmColumn].values if algorithmColumn else result.iloc[:, 0].values
        xLabel = titles[0]
        colors = ['red', 'blue', 'green', 'black', 'yellow']
        latexCode = ""
        latexCode += "\\begin{axis}[\n\txlabel={\\Huge{" + xLabel + "}},"
        latexCode += "\n\tylabel={\\Huge{" + titles[1] + "}},"
        latexCode += "\n\txmin=" + str(min(xaxisValues)) + ", xmax=" + str(max(xaxisValues)) + ",]\n"

        for num, legend in enumerate(legendary):
            color = colors[num % len(colors)]
            latexCode += "\\addplot+  [" + color + "]\n\tcoordinates {\n"
            for x, y, a in zip(xaxisValues, yaxisValues, algo):
                if legend == a:
                    latexCode += f"({x},{y})\n"
            latexCode += "\t};   \\addlegendentry{" + str(legend) + "}\n"
        latexCode += "\\end{axis}"
        self.latexCode = latexCode

    def print(self, result: pd.DataFrame, xaxis, yaxis, label=None):
        """
        Print the LaTeX code with user-specified parameters.
        """
        self.generateLatexCode(result, xaxis, yaxis, label)
        print(self.latexCode)

    def save(self, result: pd.DataFrame, xaxis, yaxis, label=None, oFile="output.tex"):
        """
        Save the generated LaTeX code to a file with user-specified parameters.
        """
        self.generateLatexCode(result, xaxis, yaxis, label)
        with open(oFile, "w") as latexWriter:
            latexWriter.write(self.latexCode)
        print(f"LaTeX file saved as {oFile}")



"""data = {
    'algorithm': ['Alg1', 'Alg2', 'Alg3', 'Alg4'],
    'minSup': [200, 400, 600, 800],
    'patterns': [800, 600, 400, 200],
    'runtime': [107.7, 98.6, 86.8, 12.5],
    'memoryRSS': [16, 15, 14, 13],
    'memoryUSS': [26, 25, 24, 23]
}
resultsDF = pd.DataFrame(data)
obj = results2Latex()

obj.print(resultsDF, xaxis='minSup', yaxis='patterns', label='algorithm')
obj.print(resultsDF, xaxis='minSup', yaxis='runtime', label='algorithm')
obj.print(resultsDF, xaxis='minSup', yaxis='memoryRSS', label='algorithm')
obj.print(resultsDF, xaxis='minSup', yaxis='memoryUSS', label='algorithm')

obj.save(resultsDF, xaxis='minSup', yaxis='patterns', label='algorithm', oFile='patterns.tex')
obj.save(resultsDF, xaxis='minSup', yaxis='runtime', label='algorithm', oFile='runtime.tex')
obj.save(resultsDF, xaxis='minSup', yaxis='memoryRSS', label='algorithm', oFile='memoryRSS.tex')
obj.save(resultsDF, xaxis='minSup', yaxis='memoryUSS', label='algorithm', oFile='memoryUSS.tex')"""



